from typing import Union

from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.erf import ERF, ERFBinaryReader, ERFBinaryWriter
from pykotor.resource.type import ResourceType


def read_erf(
        source: Union[str, bytes, bytearray, BinaryReader],
        offset: int = 0,
        size: int = None
) -> ERF:
    """
    Returns an ERF instance from the source. The file format (ERF or MOD) is automatically determined before parsing
    the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.
        size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

    Raises:
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.
        ValueError: If the file was corrupted.

    Returns:
        An ERF instance.
    """
    return ERFBinaryReader(source, offset).load()


def write_erf(
        erf: ERF,
        target: Union[str, bytearray, BinaryReader],
        file_format: ResourceType = ResourceType.ERF
) -> None:
    """
    Writes the ERF data to the target location with the specified format (ERF or MOD).

    Args:
        erf: The ERF file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be written to the specified destination.
        ValueError: If the specified format was unsupported.
    """
    if file_format == ResourceType.ERF or file_format == ResourceType.MOD:
        ERFBinaryWriter(erf, target).write()
    else:
        raise ValueError("Unsupported format specified; use ERF or MOD.")


def bytes_erf(
        erf: ERF,
        file_format: ResourceType = ResourceType.ERF
) -> bytes:
    """
    Returns the ERF data in the specified format (ERF or MOD) as a bytes object.

    This is a convience method that wraps the write_erf() method.

    Args:
        erf: The target ERF object.
        file_format: The file format.

    Raises:
        ValueError: If the specified format was unsupported.

    Returns:
        The ERF data.
    """
    data = bytearray()
    write_erf(erf, data, file_format)
    return data



from __future__ import annotations

from pykotor.common.language import LocalizedString
from pykotor.common.misc import Game, ResRef
from pykotor.resource.formats.gff import GFF, GFFContent, read_gff, write_gff
from pykotor.resource.formats.gff.gff_auto import bytes_gff
from pykotor.resource.type import ResourceType, SOURCE_TYPES, TARGET_TYPES


class UTW:
    """
    Stores waypoint data.

    resref: "TemplateResRef" field.
    tag: "Tag" field.
    name: "LocalizedName" field.
    has_map_note: "HasMapNote" field.
    map_note: "MapNote" field.
    map_note_enabled: "MapNoteEnabled" field.

    palette_id: "PaletteID" field. Used in toolset use only.
    comment: "Comment" field. Used in toolset only.
    appearance_id: "Appearance" field. Used in toolset use only.

    linked_to: "LinkedTo" field. Not used by the game engine.
    description: "Description" field. Not used by the game engine.
    """

    BINARY_TYPE = ResourceType.UTW

    def __init__(
            self
    ):
        self.resref: ResRef = ResRef.from_blank()
        self.comment: str = ""
        self.tag: str = ""

        self.name: LocalizedString = LocalizedString.from_invalid()

        self.has_map_note: bool = False
        self.map_note_enabled: bool = False
        self.map_note: LocalizedString = LocalizedString.from_invalid()

        self.appearance_id: int = 0
        self.palette_id: int = 0

        # Deprecated:
        self.linked_to: str = ""
        self.description: LocalizedString = LocalizedString.from_invalid()


def construct_utw(
        gff: GFF
) -> UTW:
    utw = UTW()

    root = gff.root
    utw.appearance_id = root.acquire("Appearance", 0)
    utw.linked_to = root.acquire("LinkedTo", "")
    utw.resref = root.acquire("TemplateResRef", ResRef.from_blank())
    utw.tag = root.acquire("Tag", "")
    utw.name = root.acquire("LocalizedName", LocalizedString.from_invalid())
    utw.description = root.acquire("Description", LocalizedString.from_invalid())
    utw.has_map_note = root.acquire("HasMapNote", 0)
    utw.map_note = root.acquire("MapNote", LocalizedString.from_invalid())
    utw.map_note_enabled = root.acquire("MapNoteEnabled", 0)
    utw.palette_id = root.acquire("PaletteID", 0)
    utw.comment = root.acquire("Comment", "")

    return utw


def dismantle_utw(
        utw: UTW,
        game: Game = Game.K2,
        *,
        use_deprecated: bool = True
) -> GFF:
    gff = GFF(GFFContent.UTW)

    root = gff.root
    root.set_uint8("Appearance", utw.appearance_id)
    root.set_string("LinkedTo", utw.linked_to)
    root.set_resref("TemplateResRef", utw.resref)
    root.set_string("Tag", utw.tag)
    root.set_locstring("LocalizedName", utw.name)
    root.set_locstring("Description", utw.description)
    root.set_uint8("HasMapNote", utw.has_map_note)
    root.set_locstring("MapNote", utw.map_note)
    root.set_uint8("MapNoteEnabled", utw.map_note_enabled)
    root.set_uint8("PaletteID", utw.palette_id)
    root.set_string("Comment", utw.comment)

    return gff


def read_utw(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> UTW:
    gff = read_gff(source, offset, size)
    utw = construct_utw(gff)
    return utw


def write_utw(
        utw: UTW,
        target: TARGET_TYPES,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> None:
    gff = dismantle_utw(utw, game, use_deprecated=use_deprecated)
    write_gff(gff, target, file_format)


def bytes_utw(
        utw: UTW,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> bytes:
    gff = dismantle_utw(utw, game, use_deprecated=use_deprecated)
    return bytes_gff(gff, file_format)



from pykotor.resource.formats.lip.lip_data import LIP, LIPShape, LIPKeyFrame
from pykotor.resource.formats.lip.io_lip import LIPBinaryReader, LIPBinaryWriter
from pykotor.resource.formats.lip.io_lip_xml import LIPXMLReader, LIPXMLWriter
from pykotor.resource.formats.lip.lip_auto import read_lip, write_lip, detect_lip



from unittest import TestCase

from pykotor.resource.formats.gff import read_gff
from pykotor.resource.generics.uti import construct_uti, dismantle_uti, UTI

TEST_FILE = "../../files/test.uti"


class TestUTI(TestCase):
    def test_io(self):
        gff = read_gff(TEST_FILE)
        uti = construct_uti(gff)
        self.validate_io(uti)

        gff = dismantle_uti(uti)
        uti = construct_uti(gff)
        self.validate_io(uti)

    def validate_io(self, uti: UTI):
        self.assertEqual("g_a_class4001", uti.resref)
        self.assertEqual(38, uti.base_item)
        self.assertEqual(5632, uti.name.stringref)
        self.assertEqual(5633, uti.description.stringref)
        self.assertEqual("G_A_CLASS4001", uti.tag)
        self.assertEqual(13, uti.charges)
        self.assertEqual(50, uti.cost)
        self.assertEqual(1, uti.stolen)
        self.assertEqual(1, uti.stack_size)
        self.assertEqual(1, uti.plot)
        self.assertEqual(50, uti.add_cost)
        self.assertEqual(1, uti.texture_variation)
        self.assertEqual(2, uti.model_variation)
        self.assertEqual(3, uti.body_variation)
        self.assertEqual(1, uti.texture_variation)
        self.assertEqual(1, uti.palette_id)
        self.assertEqual("itemo", uti.comment)

        self.assertEqual(2, len(uti.properties))
        self.assertIsNone(uti.properties[0].upgrade_type, None)
        self.assertEqual(100, uti.properties[1].chance_appear)
        self.assertEqual(1, uti.properties[1].cost_table)
        self.assertEqual(1, uti.properties[1].cost_value)
        self.assertEqual(255, uti.properties[1].param1)
        self.assertEqual(1, uti.properties[1].param1_value)
        self.assertEqual(45, uti.properties[1].property_name)
        self.assertEqual(6, uti.properties[1].subtype)
        self.assertEqual(24, uti.properties[1].upgrade_type)



import struct
from typing import List, Optional

import glm
from glm import mat4, vec3, vec4
from pykotor.common.stream import BinaryReader

from pykotor.gl.models.mdl import Node, Mesh, Model


def _load_node(scene, node: Optional[Node], mdl: BinaryReader, mdx: BinaryReader, offset: int, names: List[str]) -> Node:
    mdl.seek(offset)
    node_type = mdl.read_uint16()
    supernode_id = mdl.read_uint16()
    name_id = mdl.read_uint16()
    node = Node(scene, node, names[name_id])

    mdl.seek(offset + 16)
    node._position = glm.vec3(mdl.read_single(), mdl.read_single(), mdl.read_single())
    node._rotation = glm.quat(mdl.read_single(), mdl.read_single(), mdl.read_single(), mdl.read_single())
    node._recalc_transform()
    child_offsets = mdl.read_uint32()
    child_count = mdl.read_uint32()

    walkmesh = bool(node_type & 0b1000000000)

    if node_type & 0b100000:
        mdl.seek(offset + 80)
        fp = mdl.read_uint32()
        k2 = True if fp == 4216880 or fp == 4216816 or fp == 4216864 else False

        mdl.seek(offset + 80 + 8)
        offset_to_faces = mdl.read_uint32()
        face_count = mdl.read_uint32()

        mdl.seek(offset + 80 + 88)
        texture = mdl.read_string(32)
        lightmap = mdl.read_string(32)

        mdl.seek(offset + 80 + 313)
        node.render = bool(mdl.read_uint8())

        mdl.seek(offset + 80 + 304)
        vertex_count = mdl.read_uint16()
        if k2:
            mdl.seek(offset + 80 + 332)
        else:
            mdl.seek(offset + 80 + 324)
        mdx_offset = mdl.read_uint32()
        offset_to_vertices = mdl.read_uint32()

        element_data = []
        mdl.seek(offset + 80 + 184)
        element_offsets_count = mdl.read_uint32()
        offset_to_element_offsets = mdl.read_int32()
        if offset_to_element_offsets != -1 and element_offsets_count > 0:
            mdl.seek(offset_to_element_offsets)
            offset_to_elements = mdl.read_uint32()
            mdl.seek(offset_to_elements)
            element_data = mdl.read_bytes(face_count * 2 * 3)

        mdl.seek(offset + 80 + 252)
        mdx_block_size = mdl.read_uint32()
        mdx_data_bitflags = mdl.read_uint32()
        mdx_vertex_offset = mdl.read_int32()
        mdx_normal_offset = mdl.read_int32()
        mdl.skip(4)
        mdx_texture_offset = mdl.read_int32()
        mdx_lightmap_offset = mdl.read_int32()

        mdl.seek(offset + 80 + 313)
        render = mdl.read_uint8()

        if render and not walkmesh and element_offsets_count > 0:
            mdx.seek(mdx_offset)
            vertex_data = mdx.read_bytes(mdx_block_size * vertex_count)
            node.mesh = Mesh(scene, node, texture, lightmap, vertex_data, element_data, mdx_block_size, mdx_data_bitflags,
                             mdx_vertex_offset, mdx_normal_offset, mdx_texture_offset, mdx_lightmap_offset)

    for i in range(child_count):
        mdl.seek(child_offsets + i * 4)
        offset_to_child = mdl.read_uint32()
        new_node = _load_node(scene, node, mdl, mdx, offset_to_child, names)
        node.children.append(new_node)

    return node


def gl_load_mdl(scene, mdl: BinaryReader, mdx: BinaryReader) -> Model:
    mdl.seek(40)
    offset = mdl.read_uint32()

    mdl.seek(184)
    offset_to_name_offsets = mdl.read_uint32()
    name_count = mdl.read_uint32()

    name_offsets = []
    mdl.seek(offset_to_name_offsets)
    for i in range(name_count):
        name_offsets.append(mdl.read_uint32())

    names = []
    for name_offset in name_offsets:
        mdl.seek(name_offset)
        names.append(mdl.read_terminated_string("\0"))

    model = Model(scene, _load_node(scene, None, mdl, mdx, offset, names))
    return model


def gl_load_stitched_model(scene, mdl: BinaryReader, mdx: BinaryReader) -> Model:
    """
    Returns a model instance that has meshes with the same textures merged together.
    """
    root = Node(scene, None, "root")

    mdl.seek(40)
    offset = mdl.read_uint32()

    mdl.seek(184)
    offset_to_name_offsets = mdl.read_uint32()
    name_count = mdl.read_uint32()

    name_offsets = []
    mdl.seek(offset_to_name_offsets)
    for i in range(name_count):
        name_offsets.append(mdl.read_uint32())

    names = []
    for name_offset in name_offsets:
        mdl.seek(name_offset)
        names.append(mdl.read_terminated_string("\0"))

    # Build list offset to nodes that: 1. Have meshes 2. Will render ingame
    offsets = []
    search = [(offset, mat4())]
    while search:
        offset, transform = search.pop()

        mdl.seek(offset)
        node_type = mdl.read_uint16()
        supernode_id = mdl.read_uint16()
        name_id = mdl.read_uint16()

        mdl.seek(offset + 16)
        position = glm.vec3(mdl.read_single(), mdl.read_single(), mdl.read_single())
        rotation = glm.quat(mdl.read_single(), mdl.read_single(), mdl.read_single(), mdl.read_single())
        child_offsets = mdl.read_uint32()
        child_count = mdl.read_uint32()

        transform = transform * glm.translate(position) * glm.mat4_cast(rotation)

        for i in range(child_count):
            mdl.seek(child_offsets + i * 4)
            offset_to_child = mdl.read_uint32()
            search.append((offset_to_child, transform))

        walkmesh = bool(node_type & 0b1000000000)
        mesh = bool(node_type & 0b100000)

        if mesh and not walkmesh:
            mdl.seek(offset + 80 + 313)
            render = mdl.read_uint8()
            if render:
                offsets.append((offset, transform))

        if names[name_id].lower() in ["headhook", "rhand", "lhand", "gogglehook", "maskhook"]:
            node = Node(scene, root, names[name_id])
            root.children.append(node)
            glm.decompose(transform, vec3(), node._rotation, node._position, vec3(), vec4())
            node._recalc_transform()

    merged = {}
    for offset, transform in offsets:
        mdl.seek(offset + 80 + 88)
        texture = mdl.read_string(32)
        lightmap = mdl.read_string(32)
        key = texture + "\n" + lightmap
        if key not in merged:
            merged[key] = []
        merged[key].append((offset, transform))

    for key in merged:
        vertex_data = bytearray()
        elements = []
        child = Node(scene, root, "child")
        root.children.append(child)

        last_element = 0
        for offset, transform in merged[key]:
            mdl.seek(offset + 80)
            fp = mdl.read_uint32()
            k2 = True if fp == 4216880 or fp == 4216816 or fp == 4216864 else False

            mdl.seek(offset + 80 + 252)
            mdx_block_size = mdl.read_uint32()
            mdx_data_bitflags = mdl.read_uint32()
            mdx_vertex_offset = mdl.read_int32()
            mdx_normal_offset = mdl.read_int32()
            mdl.skip(4)
            mdx_texture_offset = mdl.read_int32()
            mdx_lightmap_offset = mdl.read_int32()

            mdl.seek(offset + 80 + 8)
            offset_to_faces = mdl.read_uint32()
            face_count = mdl.read_uint32()
            mdl.seek(offset + 80 + 184)
            element_offsets_count = mdl.read_uint32()
            offset_to_element_offsets = mdl.read_int32()
            if offset_to_element_offsets != -1 and element_offsets_count > 0:
                mdl.seek(offset_to_element_offsets)
                offset_to_elements = mdl.read_uint32()
                mdl.seek(offset_to_elements)
                [elements.append(mdl.read_uint16() + last_element) for i in range(face_count * 3)]

            mdl.seek(offset + 80 + 304)
            vertex_count = mdl.read_uint16()
            if k2:
                mdl.seek(offset + 80 + 332)
            else:
                mdl.seek(offset + 80 + 324)
            mdx_offset = mdl.read_uint32()

            for i in range(vertex_count):
                mdx.seek(mdx_offset + i * mdx_block_size + mdx_vertex_offset)
                vertex = vec3(mdx.read_single(), mdx.read_single(), mdx.read_single())
                vertex = transform * vertex
                vertex_data += struct.pack('fff', vertex.x, vertex.y, vertex.z)

                if mdx_normal_offset == -1:
                    vertex_data += bytes(12)
                else:
                    mdx.seek(mdx_offset + i * mdx_block_size + mdx_normal_offset)
                    vertex_data += mdx.read_bytes(12)

                if mdx_texture_offset == -1:
                    vertex_data += bytes(8)
                else:
                    mdx.seek(mdx_offset + i * mdx_block_size + mdx_texture_offset)
                    vertex_data += mdx.read_bytes(8)

                if mdx_lightmap_offset == -1:
                    vertex_data += bytes(8)
                else:
                    mdx.seek(mdx_offset + i * mdx_block_size + mdx_lightmap_offset)
                    vertex_data += mdx.read_bytes(8)

            last_element += vertex_count

        element_data = bytearray()
        for element in elements:
            element_data += struct.pack('H', element)

        texture, lightmap = key.split("\n")
        child.mesh = Mesh(scene, child, texture, lightmap, vertex_data, element_data, 40, mdx_data_bitflags, 0, 12, 24, 32)

    return Model(scene, root)



from pykotor.resource.formats.mdl.mdl_data import MDL, MDLNode, MDLMesh
from pykotor.resource.formats.mdl.io_mdl import MDLBinaryReader, MDLBinaryWriter
from pykotor.resource.formats.mdl.io_mdl_ascii import MDLAsciiReader, MDLAsciiWriter
from pykotor.resource.formats.mdl.mdl_auto import write_mdl, read_mdl



import platform
from unittest import TestCase

from pykotor.resource.formats.twoda import write_2da, TwoDABinaryReader, read_2da, detect_2da, TwoDACSVReader, TwoDA
from pykotor.resource.formats.twoda.io_twoda_json import TwoDAJSONReader
from pykotor.resource.type import ResourceType


BINARY_TEST_FILE = "../../files/test.2da"
CSV_TEST_FILE = "../../files/test.2da.csv"
JSON_TEST_FILE = "../../files/test.2da.json"
DOES_NOT_EXIST_FILE = "./thisfiledoesnotexist"
CORRUPT_BINARY_TEST_FILE = "../../files/test_corrupted.2da"
CORRUPT_CSV_TEST_FILE = "../../files/test_corrupted.2da.csv"
CORRUPT_JSON_TEST_FILE = "../../files/test_corrupted.2da.json"


class TestTwoDA(TestCase):
    def test_binary_io(self):
        self.assertEqual(detect_2da(BINARY_TEST_FILE), ResourceType.TwoDA)
        twoda = TwoDABinaryReader(BINARY_TEST_FILE).load()
        self.validate_io(twoda)

        data = bytearray()
        write_2da(twoda, data, ResourceType.TwoDA)
        twoda = read_2da(data)
        self.validate_io(twoda)

    def test_csv_io(self):
        self.assertEqual(detect_2da(CSV_TEST_FILE), ResourceType.TwoDA_CSV)

        twoda = TwoDACSVReader(CSV_TEST_FILE).load()
        self.validate_io(twoda)

        data = bytearray()
        write_2da(twoda, data, ResourceType.TwoDA_CSV)
        twoda = read_2da(data)
        self.validate_io(twoda)

    def test_json_io(self):
        self.assertEqual(detect_2da(JSON_TEST_FILE), ResourceType.TwoDA_JSON)

        twoda = TwoDAJSONReader(JSON_TEST_FILE).load()
        self.validate_io(twoda)

        data = bytearray()
        write_2da(twoda, data, ResourceType.TwoDA_JSON)
        twoda = read_2da(data)
        self.validate_io(twoda)

    def validate_io(self, twoda):
        self.assertEqual("abc", twoda.get_cell(0, "col1"))
        self.assertEqual("def", twoda.get_cell(0, "col2"))
        self.assertEqual("ghi", twoda.get_cell(0, "col3"))

        self.assertEqual("def", twoda.get_cell(1, "col1"))
        self.assertEqual("ghi", twoda.get_cell(1, "col2"))
        self.assertEqual("123", twoda.get_cell(1, "col3"))

        self.assertEqual("123", twoda.get_cell(2, "col1"))
        self.assertEqual("", twoda.get_cell(2, "col2"))
        self.assertEqual("abc", twoda.get_cell(2, "col3"))

    def test_read_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, read_2da, ".")
        else:
            self.assertRaises(IsADirectoryError, read_2da, ".")
        self.assertRaises(FileNotFoundError, read_2da, DOES_NOT_EXIST_FILE)
        self.assertRaises(ValueError, read_2da, CORRUPT_BINARY_TEST_FILE)
        self.assertRaises(ValueError, read_2da, CORRUPT_CSV_TEST_FILE)
        self.assertRaises(ValueError, read_2da, CORRUPT_JSON_TEST_FILE)

    def test_write_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, write_2da, TwoDA(), ".", ResourceType.TwoDA)
        else:
            self.assertRaises(IsADirectoryError, write_2da, TwoDA(), ".", ResourceType.TwoDA)
        self.assertRaises(ValueError, write_2da, TwoDA(), ".", ResourceType.INVALID)

    def test_row_max(self):
        twoda = TwoDA()
        twoda.add_row("0")
        twoda.add_row("1")
        twoda.add_row("2")

        self.assertEqual("3", twoda.label_max())



import os
from time import sleep

from ply.lex import LexError

from pykotor.common.script import DataType
from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.ncs import NCS, write_ncs, read_ncs
from pykotor.resource.formats.ncs.compiler.interpreter import Interpreter, Stack
from pykotor.resource.formats.ncs.compiler.lexer import NssLexer
from pykotor.resource.formats.ncs.compiler.parser import NssParser
from pykotor.resource.formats.ncs.optimizers import RemoveNopOptimizer


def main1():
    nssLexer = NssLexer()
    nssPareser = NssParser()

    lex = nssLexer.lexer
    parser = nssPareser.parser

    aaa = """
    int num = 123;
    string tag = "abc";
    object a = GetObjectByTag(tag, num);
    """

    bbb = """
    int num = GetPartyMemberCount();
    object a = GetObjectByTag(GetStringByStrRef(0), num);
    """

    ccc = """
    void main() {
        int abc = 0;
        if (abc)
        {
            abc = 1;
        }
        abc = 2;
    }
    """

    # EngineCalls can be called without declaration/assignment
    # but still adds to stack unrecorderd
    # need to fix.

    t = parser.parse(ccc, tracking=True)

    #GetObjectByTag(tag, num);

    ncs = NCS()
    t.compile(ncs)
    for inst in ncs.instructions:
        print(inst)


def compile(script):
    nssLexer = NssLexer()
    nssPareser = NssParser()

    lex = nssLexer.lexer
    parser = nssPareser.parser

    ncs = NCS()
    t = parser.parse(script)
    t.compile(ncs)
    return ncs


def main2():
    dir1 = "C:/Users/hugin/Desktop/Scripts/K1/"
    dir2 = "C:/Users/hugin/Desktop/Scripts/K2/"

    success = []
    failed = []

    for file in os.listdir(dir1):
        try:
            script = BinaryReader.load_file(dir1 + file).decode()
            compile(script)
            success.append(file)
        except (Exception, LexError) as e:
            failed.append(file)
    for file in os.listdir(dir2):
        try:
            script = BinaryReader.load_file(dir2 + file).decode()
            compile(script)
            success.append(file)
        except (Exception, LexError) as e:
            failed.append(file)

    total_scripts = len(success) + len(failed)
    percentage = int(len(success)/total_scripts*100)
    print("{}/{} ({}%)".format(len(success), total_scripts, percentage))
    print(success)


def main4():
    dir1 = "C:/Users/hugin/Desktop/Scripts/K1/"
    dir2 = "C:/Users/hugin/Desktop/Scripts/K2/"

    for file in os.listdir(dir1):
        script = BinaryReader.load_file(dir1 + file).decode(errors="ignore")
        if "for(" in script or "for (" in script:
            print(file)


def main3():
    script = """
    
    void main()
    {
        int a = 5;
        int b = 0;
        b += 2 * 2;
    }
    """
    ncs = compile(script)
    #ncs.print()

    interpreter = Interpreter(ncs)
    interpreter.run()
    interpreter.print()


def main5():
    script = """
            #include "k_inc_cheat"

            void main()
            {
                case 1:
                
            }
        """
    ncs = compile(script)
    ncs.optimize([RemoveNopOptimizer()])
    ncs.print()
    write_ncs(ncs, "C:/Users/hugin/Desktop/ext/epic.ncs")

    interpreter = Interpreter(ncs)
    interpreter.run()


def main6():
    ncs = read_ncs("C:/Users/hugin/Desktop/ext/epic.ncs")
    #ncs.print()

    interpreter = Interpreter(ncs)
    interpreter.run()

    interpreter.print()
    print(interpreter.action_snapshots)



def main7():
    stack = Stack()
    stack.add(DataType.FLOAT, 1)  # -24
    stack.add(DataType.FLOAT, 2)  # -20
    stack.add(DataType.FLOAT, 3)  # -17
    stack.add(DataType.FLOAT, 4)  # -12
    stack.add(DataType.FLOAT, 5)  # -8
    stack.add(DataType.FLOAT, 6)  # -4
    stack.copy_to_top(-12, 4)


main6()



"""
This module handles classes relating to editing ERF files.
"""
from __future__ import annotations

from copy import copy
from enum import Enum
from typing import List, Optional

from pykotor.common.misc import ResRef
from pykotor.resource.type import ResourceType


class ERFType(Enum):
    """
    The type of ERF.
    """
    ERF = "ERF "
    MOD = "MOD "

    @staticmethod
    def from_extension(filepath: str) -> ERFType:
        if filepath.lower().endswith(".erf"):
            return ERFType.ERF
        elif filepath.lower().endswith(".mod"):
            return ERFType.MOD
        else:
            raise ValueError(f"Invalid ERF extension in filepath '{filepath}'.")


class ERF:
    """
    Represents the data of a ERF file.

    Attributes:
        erf_type: The ERF type.
    """

    BINARY_TYPE = ResourceType.ERF

    def __init__(
            self,
            erf_type: ERFType = ERFType.ERF
    ):
        self.erf_type: ERFType = erf_type
        self._resources: List[ERFResource] = []

    def __iter__(
            self
    ):
        """
        Iterates through the stored resources yielding a copied resource each iteration.
        """
        for resource in self._resources:
            yield copy(resource)

    def __len__(
            self
    ):
        """
        Returns the number of stored resources.
        """
        return len(self._resources)

    def __getitem__(
            self,
            item
    ):
        """
        Returns a resource at the specified index or with the specified resref.
        """
        if isinstance(item, int):
            return self._resources[item]
        elif isinstance(item, str):
            resource = next([resource for resource in self._resources if resource.resref == item], None)
            if resource is None:
                raise KeyError
            else:
                return resource
        else:
            return NotImplemented

    def set(
            self,
            resref: str,
            restype: ResourceType,
            data: bytes
    ) -> None:
        """
        Sets the data of the resource with the specified resref/restype pair. If it does not exists, a resource is
        appended to the resource list.

        Args:
            resref: The resref.
            restype: The resource type.
            data: The new resource data.
        """
        resource = next(
            (resource for resource in self._resources if resource.resref == resref and resource.restype == restype),
            None)
        if resource is None:
            self._resources.append(ERFResource(ResRef(resref), restype, data))
        else:
            resource.resref = ResRef(resref)
            resource.restype = restype
            resource.data = data

    def get(
            self,
            resref: str,
            restype: ResourceType
    ) -> Optional[bytes]:
        """
        Returns the data of the resource with the specified resref/restype pair if it exists, otherwise returns None.

        Args:
            resref: The resref.
            restype: The resource type.

        Returns:
            The bytes data of the resource or None.
        """
        resource = next(
            (resource for resource in self._resources if resource.resref == resref and resource.restype == restype),
            None)
        return None if resource is None else resource.data

    def remove(
            self,
            resref: str,
            restype: ResourceType
    ) -> None:
        """
        Removes the resource with the given resref/restype pair if it exists.

        Args:
            resref: The resref.
            restype: The resource type.
        """
        self._resources = [res for res in self._resources if res.resref != resref and res.restype != restype]

    def to_rim(
            self
    ):
        """
        Returns a RIM with the same resources.

        Returns:
            A new RIM object.
        """
        from pykotor.resource.formats.rim import RIM  # Prevent circular imports
        rim = RIM()
        for resource in self._resources:
            rim.set(resource.resref.get(), resource.restype, resource.data)
        return rim


class ERFResource:
    def __init__(
            self,
            resref: ResRef,
            restype: ResourceType,
            data: bytes
    ):
        self.resref: ResRef = resref
        self.restype: ResourceType = restype
        self.data: bytes = data



from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.tpc import TPC, TPCBinaryReader, TPCBinaryWriter, TPCTGAWriter, TPCBMPWriter, TPCTGAReader
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceType


def detect_tpc(
        source: SOURCE_TYPES,
        offset: int = 0
) -> ResourceType:
    """
    Returns what format the TPC data is believed to be in. This function performs a basic check and does not guarantee
    accuracy of the result or integrity of the data.

    Args:
        source: Source of the TPC data.
        offset: Offset into the source data.

    Raises:
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.

    Returns:
        The format of the TPC data.
    """

    def do_check(
            first100: bytes
    ) -> ResourceType:
        file_format = ResourceType.TPC
        if len(first100) < 100:
            file_format = ResourceType.TGA
        else:
            for i in range(15, 100):
                if first100[i] != 0:
                    file_format = ResourceType.TGA
        return file_format

    try:
        if isinstance(source, str):
            with BinaryReader.from_file(source, offset) as reader:
                file_format = do_check(reader.read_bytes(100))
        elif isinstance(source, bytes) or isinstance(source, bytearray):
            file_format = do_check(source[:100])
        elif isinstance(source, BinaryReader):
            file_format = do_check(source.read_bytes(100))
            source.skip(-100)
        else:
            file_format = ResourceType.INVALID
    except (FileNotFoundError, PermissionError, IsADirectoryError) as e:
        raise e
    except IOError:
        file_format = ResourceType.INVALID

    return file_format


def read_tpc(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> TPC:
    """
    Returns an TPC instance from the source. The file format (TPC or TGA) is automatically determined before
    parsing the data.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.
        size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

    Raises:
        FileNotFoundError: If the file could not be found.
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be accessed.
        ValueError: If the file was corrupted or the format could not be determined.

    Returns:
        An TPC instance.
    """
    file_format = detect_tpc(source, offset)

    if file_format is ResourceType.INVALID:
        raise ValueError("Failed to determine the format of the GFF file.")

    if file_format == ResourceType.TPC:
        return TPCBinaryReader(source, offset, size).load()
    elif file_format == ResourceType.TGA:
        return TPCTGAReader(source, offset, size).load()


def write_tpc(
        tpc: TPC,
        target: TARGET_TYPES,
        file_format: ResourceType = ResourceType.TPC
) -> None:
    """
    Writes the TPC data to the target location with the specified format (TPC, TGA or BMP).

    Args:
        tpc: The TPC file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        IsADirectoryError: If the specified path is a directory (Unix-like systems only).
        PermissionError: If the file could not be written to the specified destination.
        ValueError: If the specified format was unsupported.
    """
    if file_format == ResourceType.TGA:
        TPCTGAWriter(tpc, target).write()
    elif file_format == ResourceType.BMP:
        TPCBMPWriter(tpc, target).write()
    elif file_format == ResourceType.TPC:
        TPCBinaryWriter(tpc, target).write()
    else:
        raise ValueError("Unsupported format specified; use TPC, TGA or BMP.")


def bytes_tpc(
        tpc: TPC,
        file_format: ResourceType = ResourceType.TPC
) -> bytes:
    """
    Returns the TPC data in the specified format (TPC, TGA or BMP) as a bytes object.

    This is a convience method that wraps the write_tpc() method.

    Args:
        tpc: The target TPC object.
        file_format: The file format.

    Raises:
        ValueError: If the specified format was unsupported.

    Returns:
        The TPC data.
    """
    data = bytearray()
    write_tpc(tpc, data, file_format)
    return data



from unittest import TestCase

from pykotor.common.stream import BinaryReader


class TestBinaryReader(TestCase):
    def setUp(self) -> None:
        self.data1 = b'\x01' + b'\x02\x00' + b'\x03\x00\x00\x00' + b'\x04\x00\x00\x00\x00\x00\x00\x00'
        self.data2 = b'helloworld\x00'
        self.data3 = b'\xFF' + b'\xFE\xFF' + b'\xFD\xFF\xFF\xFF' + b'\xFC\xFF\xFF\xFF\xFF\xFF\xFF\xFF'
        self.data4 = b'\x79\xE9\xF6\xC2' + b'\x68\x91\xED\x7C\x3F\xDD\x5E\x40'

        self.reader1 = BinaryReader.from_bytes(self.data1)
        self.reader1b = BinaryReader.from_bytes(self.data1, 3)
        self.reader1c = BinaryReader.from_bytes(self.data1, 3, 4)
        self.reader2 = BinaryReader.from_bytes(self.data2)
        self.reader3 = BinaryReader.from_bytes(self.data3)
        self.reader4 = BinaryReader.from_bytes(self.data4)

    def test_read(self):
        self.assertEqual(1, self.reader1.read_uint8())
        self.assertEqual(2, self.reader1.read_uint16())
        self.assertEqual(3, self.reader1.read_uint32())
        self.assertEqual(4, self.reader1.read_uint64())

        self.assertEqual(3, self.reader1b.read_uint32())
        self.assertEqual(4, self.reader1b.read_uint64())

        reader2 = BinaryReader.from_bytes(self.data2)
        self.assertEqual("helloworld", reader2.read_string(10))

        reader3 = BinaryReader.from_bytes(self.data3)
        self.assertEqual(-1, reader3.read_int8())
        self.assertEqual(-2, reader3.read_int16())
        self.assertEqual(-3, reader3.read_int32())
        self.assertEqual(-4, reader3.read_int64())

        reader4 = BinaryReader.from_bytes(self.data4)
        self.assertAlmostEqual(-123.456, reader4.read_single(), 3)
        self.assertAlmostEqual(--123.457, reader4.read_double(), 6)

    def test_size(self):
        self.reader1.read_bytes(4)
        self.assertEqual(15, self.reader1.size())

        self.reader1b.read_bytes(4)
        self.assertEqual(12, self.reader1b.size())

        self.reader1c.read_bytes(1)
        self.assertEqual(4, self.reader1c.size())

    def test_true_size(self):
        self.reader1.read_bytes(4)
        self.assertEqual(15, self.reader1.true_size())

        self.reader1b.read_bytes(4)
        self.assertEqual(15, self.reader1b.true_size())

        self.reader1c.read_bytes(4)
        self.assertEqual(15, self.reader1c.true_size())

    def test_position(self):
        self.reader1.read_bytes(3)
        self.reader1.read_bytes(3)
        self.assertEqual(6, self.reader1.position())

        self.reader1b.read_bytes(1)
        self.reader1b.read_bytes(2)
        self.assertEqual(3, self.reader1b.position())

        self.reader1c.read_bytes(1)
        self.reader1c.read_bytes(2)
        self.assertEqual(3, self.reader1c.position())

    def test_seek(self):
        self.reader1.read_bytes(4)
        self.reader1.seek(7)
        self.assertEqual(7, self.reader1.position())
        self.assertEqual(4, self.reader1.read_uint64())

        self.reader1b.read_bytes(3)
        self.reader1b.seek(4)
        self.assertEqual(4, self.reader1b.position())
        self.assertEqual(4, self.reader1b.read_uint32())

        self.reader1c.read_bytes(3)
        self.reader1c.seek(2)
        self.assertEqual(2, self.reader1c.position())
        self.assertEqual(0, self.reader1c.read_uint16())

    def test_skip(self):
        self.reader1.read_uint32()
        self.reader1.skip(2)
        self.reader1.skip(1)
        self.assertEqual(4, self.reader1.read_uint64())

        self.reader1b.skip(4)
        self.assertEqual(4, self.reader1b.read_uint64())

        self.reader1c.skip(2)
        self.assertEqual(0, self.reader1c.read_uint16())

    def test_remaining(self):
        self.reader1.read_uint32()
        self.reader1.skip(2)
        self.reader1.skip(1)
        self.assertEqual(8, self.reader1.remaining())

        self.reader1b.read_uint32()
        self.assertEqual(8, self.reader1b.remaining())

        self.reader1c.read_uint16()
        self.assertEqual(2, self.reader1c.remaining())

    def test_peek(self):
        self.reader1.skip(3)
        self.assertEqual(b'\x03', self.reader1.peek(1))

        self.reader1b.skip(4)
        self.assertEqual(b'\x04', self.reader1b.peek(1))

        self.assertEqual(b'\x03', self.reader1c.peek(1))



from __future__ import annotations

from typing import List

from pykotor.common.language import LocalizedString
from pykotor.common.misc import Game, ResRef
from pykotor.resource.formats.gff import GFF, GFFList, GFFContent, read_gff, write_gff
from pykotor.resource.formats.gff.gff_auto import bytes_gff
from pykotor.resource.type import ResourceType, SOURCE_TYPES, TARGET_TYPES


class UTS:
    """
    Stores sound data.

    Attributes:
        tag: "Tag" field.
        resref: "TemplateResRef" field.
        active: "Active" field.
        continuous: "Continuous" field.
        looping: "Looping" field.
        positional: "Positional" field.
        random_position: "RandomPosition" field.
        random_pick: "Random" field.
        elevation: "Elevation" field.
        max_distance: "MaxDistance" field.
        min_distance: "MinDistance" field.
        random_range_x: "RandomRangeX" field.
        random_range_y: "RandomRangeY" field.
        interval: "Interval" field.
        interval_variation: "IntervalVrtn" field.
        pitch_variation: "PitchVariation" field.
        priority: "Priority" field.
        volume: "Volume" field.
        volume_variation: "VolumeVrtn" field.
        comment: "Comment" field.

        palette_id: "PaletteID" field. Used in toolset only.

        name: "LocName" field. Not used by the game engine.
        hours: "Hours" field. Not used by the game engine.
        times: "Times" field. Not used by the game engine.
    """

    BINARY_TYPE = ResourceType.UTS

    def __init__(
            self
    ):
        self.resref: ResRef = ResRef.from_blank()
        self.tag: str = ""
        self.comment: str = ""

        self.active: bool = False
        self.continuous: bool = False
        self.looping: bool = False
        self.positional: bool = False
        self.random_pick: bool = False

        self.random_position: bool = False
        self.random_range_x: float = 0.0
        self.random_range_y: float = 0.0

        self.elevation: float = 0.0
        self.max_distance: float = 0.0
        self.min_distance: float = 0.0

        self.priority: int = 0

        self.interval: int = 0
        self.interval_variation: int = 0
        self.pitch_variation: float = 0.0
        self.volume: int = 0
        self.volume_variation: int = 0

        self.sounds: List[ResRef] = []

        # Deprecated:
        self.name: LocalizedString = LocalizedString.from_invalid()
        self.times: int = 0
        self.hours: int = 0
        self.palette_id: int = 0


def construct_uts(
        gff: GFF
) -> UTS:
    uts = UTS()

    root = gff.root
    uts.tag = root.acquire("Tag", "")
    uts.resref = root.acquire("TemplateResRef", ResRef.from_blank())
    uts.active = root.acquire("Active", 0)
    uts.continuous = root.acquire("Continuous", 0)
    uts.looping = root.acquire("Looping", 0)
    uts.positional = root.acquire("Positional", 0)
    uts.random_position = root.acquire("RandomPosition", 0)
    uts.random_pick = root.acquire("Random", 0)
    uts.elevation = root.acquire("Elevation", 0.0)
    uts.max_distance = root.acquire("MaxDistance", 0.0)
    uts.min_distance = root.acquire("MinDistance", 0.0)
    uts.random_range_x = root.acquire("RandomRangeX", 0.0)
    uts.random_range_y = root.acquire("RandomRangeY", 0.0)
    uts.interval = root.acquire("Interval", 0)
    uts.interval_variation = root.acquire("IntervalVrtn", 0)
    uts.pitch_variation = root.acquire("PitchVariation", 0.0)
    uts.priority = root.acquire("Priority", 0)
    uts.volume = root.acquire("Volume", 0)
    uts.volume_variation = root.acquire("VolumeVrtn", 0)
    uts.comment = root.acquire("Comment", "")
    uts.name = root.acquire("LocName", LocalizedString.from_invalid())
    uts.hours = root.acquire("Hours", 0)
    uts.times = root.acquire("Times", 0)
    uts.palette_id = root.acquire("PaletteID", 0)

    for sound_struct in root.acquire("Sounds", GFFList()):
        sound = sound_struct.acquire("Sound", ResRef.from_blank())
        uts.sounds.append(sound)

    return uts


def dismantle_uts(
        uts: UTS,
        game: Game = Game.K2,
        *,
        use_deprecated: bool = True
) -> GFF:
    gff = GFF(GFFContent.UTS)

    root = gff.root
    root.set_string("Tag", uts.tag)
    root.set_resref("TemplateResRef", uts.resref)
    root.set_uint8("Active", uts.active)
    root.set_uint8("Continuous", uts.continuous)
    root.set_uint8("Looping", uts.looping)
    root.set_uint8("Positional", uts.positional)
    root.set_uint8("RandomPosition", uts.random_position)
    root.set_uint8("Random", uts.random_pick)
    root.set_single("Elevation", uts.elevation)
    root.set_single("MaxDistance", uts.max_distance)
    root.set_single("MinDistance", uts.min_distance)
    root.set_single("RandomRangeX", uts.random_range_x)
    root.set_single("RandomRangeY", uts.random_range_y)
    root.set_uint32("Interval", uts.interval)
    root.set_uint32("IntervalVrtn", uts.interval_variation)
    root.set_single("PitchVariation", uts.pitch_variation)
    root.set_uint8("Priority", uts.priority)
    root.set_uint8("Volume", uts.volume)
    root.set_uint8("VolumeVrtn", uts.volume_variation)
    root.set_string("Comment", uts.comment)

    sound_list = root.set_list("Sounds", GFFList())
    for sound in uts.sounds:
        sound_list.add(0).set_resref("Sound", sound)

    root.set_uint8("PaletteID", uts.palette_id)

    if use_deprecated:
        root.set_locstring("LocName", uts.name)
        root.set_uint32("Hours", uts.hours)
        root.set_uint32("Times", uts.times)

    return gff


def read_uts(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> UTS:
    gff = read_gff(source, offset, size)
    uts = construct_uts(gff)
    return uts


def write_uts(
        uts: UTS,
        target: TARGET_TYPES,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> None:
    gff = dismantle_uts(uts, game, use_deprecated=use_deprecated)
    write_gff(gff, target, file_format)


def bytes_uts(
        uts: UTS,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> bytes:
    gff = dismantle_uts(uts, game, use_deprecated=use_deprecated)
    return bytes_gff(gff, file_format)



import platform
from unittest import TestCase

from pykotor.common.geometry import Vector3, Vector4
from pykotor.resource.formats.lyt import LYTAsciiReader, LYTRoom, LYTTrack, LYT, LYTObstacle, LYTDoorHook
from pykotor.resource.formats.lyt.lyt_auto import write_lyt, read_lyt
from pykotor.resource.type import ResourceType


ASCII_TEST_FILE = "../../files/test.lyt"
DOES_NOT_EXIST_FILE = "./thisfiledoesnotexist"
CORRUPT_BINARY_TEST_FILE = "../../files/test_corrupted.lyt"


class TestLYT(TestCase):
    def test_binary_io(self):
        lyt = LYTAsciiReader(ASCII_TEST_FILE).load()
        self.validate_io(lyt)

        data = bytearray()
        write_lyt(lyt, data, ResourceType.LYT)
        lyt = read_lyt(data)
        self.validate_io(lyt)

    def validate_io(self, lyt: LYT):
        self.assertEqual(lyt.rooms[0], LYTRoom("M17mg_01a", Vector3(100.0, 100.0, 0.0)))
        self.assertEqual(lyt.rooms[1], LYTRoom("M17mg_01b", Vector3(100.0, 100.0, 0.0)))
        self.assertEqual(lyt.tracks[0], LYTTrack("M17mg_MGT01", Vector3(0.0, 0.0, 0.0)))
        self.assertEqual(lyt.tracks[1], LYTTrack("M17mg_MGT02", Vector3(112.047, 209.04, 0.0)))
        self.assertEqual(lyt.obstacles[0], LYTObstacle("M17mg_MGO01", Vector3(103.309, 3691.61, 0.0)))
        self.assertEqual(lyt.obstacles[1], LYTObstacle("M17mg_MGO02", Vector3(118.969, 3688.0, 0.0)))
        self.assertEqual(lyt.doorhooks[0], LYTDoorHook("M02ac_02h", "door_01", Vector3(170.475, 66.375, 0.0),
                                                       Vector4(0.707107, 0.0, 0.0, -0.707107)))
        self.assertEqual(lyt.doorhooks[1],
                         LYTDoorHook("M02ac_02a", "door_06", Vector3(90.0, 129.525, 0.0), Vector4(1.0, 0.0, 0.0, 0.0)))
    
    def test_read_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, read_lyt, ".")
        else:
            self.assertRaises(IsADirectoryError, read_lyt, ".")
        self.assertRaises(FileNotFoundError, read_lyt, DOES_NOT_EXIST_FILE)
        self.assertRaises(ValueError, read_lyt, CORRUPT_BINARY_TEST_FILE)

    def test_write_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, write_lyt, LYT(), ".", ResourceType.LYT)
        else:
            self.assertRaises(IsADirectoryError, write_lyt, LYT(), ".", ResourceType.LYT)
        self.assertRaises(ValueError, write_lyt, LYT(), ".", ResourceType.INVALID)



from pykotor.common.misc import Game
from pykotor.common.scriptdefs import KOTOR_FUNCTIONS, TSL_FUNCTIONS, KOTOR_CONSTANTS, TSL_CONSTANTS
from pykotor.common.scriptlib import TSL_LIBRARY, KOTOR_LIBRARY
from pykotor.resource.formats.ncs import NCS, NCSBinaryReader, NCSBinaryWriter
from pykotor.resource.formats.ncs.compiler.lexer import NssLexer
from pykotor.resource.formats.ncs.compiler.parser import NssParser
from pykotor.resource.formats.ncs.optimizers import RemoveNopOptimizer, RemoveMoveSPEqualsZeroOptimizer
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceType


def read_ncs(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> NCS:
    """
    Returns an NCS instance from the source.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.
        size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

    Raises:
        ValueError: If the file was corrupted or in an unsupported format.

    Returns:
        An NCS instance.
    """
    return NCSBinaryReader(source, offset, size).load()


def write_ncs(
        ncs: NCS,
        target: TARGET_TYPES,
        file_format: ResourceType = ResourceType.NCS
) -> None:
    """
    Writes the NCS data to the target location with the specified format (NCS only).

    Args:
        ncs: The NCS file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        ValueError: If an unsupported file format was given.
    """
    if file_format == ResourceType.NCS:
        NCSBinaryWriter(ncs, target).write()
    else:
        raise ValueError("Unsupported format specified; use NCS.")


def bytes_ncs(
        ncs: NCS,
        file_format: ResourceType = ResourceType.NCS
) -> bytes:
    """
    Returns the NCS data in the specified format (NCS only) as a bytes object.

    This is a convenience method that wraps the write_ncs() method.

    Args:
        ncs: The target NCS object.
        file_format: The file format.

    Raises:
        ValueError: If an unsupported file format was given.

    Returns:
        The NCS data.
    """
    data = bytearray()
    write_ncs(ncs, data, file_format)
    return data


def compile_nss(
        source: str,
        game: Game,
        optimize: bool = False
) -> NCS:
    nssLexer = NssLexer()
    nssParser = NssParser(
        library=KOTOR_LIBRARY if game == Game.K1 else TSL_LIBRARY,
        functions=KOTOR_FUNCTIONS if game == Game.K1 else TSL_FUNCTIONS,
        constants=KOTOR_CONSTANTS if game == Game.K1 else TSL_CONSTANTS,
        library_lookup=None
    )

    ncs = NCS()

    block = nssParser.parser.parse(source, tracking=True)
    block.compile(ncs)

    ncs.optimize([RemoveNopOptimizer()])

    return ncs



import operator

import ply.lex

from pykotor.common.script import DataType
from pykotor.resource.formats.ncs import NCSInstructionType
from pykotor.resource.formats.ncs.compiler.classes import IntExpression, ControlKeyword, StringExpression, \
    FloatExpression, Identifier, Operator, BinaryOperatorMapping, UnaryOperatorMapping, OperatorMapping, \
    ObjectExpression

from abc import ABC
from enum import Enum
from typing import Any

import ply.lex as lex


class NssLexer:
    def __init__(self, errorlog=lex.NullLogger()):
        self.lexer = lex.lex(module=self, errorlog=errorlog)
        self.lexer.begin("INITIAL")

    tokens = [
        "STRING_VALUE", "INT_VALUE", "FLOAT_VALUE", "IDENTIFIER", "INT_TYPE", "INT_HEX_VALUE", "FLOAT_TYPE", "OBJECT_TYPE",
        "VOID_TYPE",  "EVENT_TYPE", "EFFECT_TYPE", "ITEMPROPERTY_TYPE", "LOCATION_TYPE", "STRING_TYPE", "TALENT_TYPE",
        "VECTOR_TYPE", "ACTION_TYPE", "BREAK_CONTROL", "CASE_CONTROL", "DEFAULT_CONTROL", "DO_CONTROL", "ELSE_CONTROL",
        "SWITCH_CONTROL", "WHILE_CONTROL", "FOR_CONTROL", "IF_CONTROL", "TRUE_VALUE", "FALSE_VALUE", "OBJECTSELF_VALUE",
        "OBJECTINVALID_VALUE", "ADD", "MINUS", "MULTIPLY", "DIVIDE",
        "MOD", "EQUALS", "NOT_EQUALS", "GREATER_THAN", "LESS_THAN",
        "LESS_THAN_OR_EQUALS", "GREATER_THAN_OR_EQUALS", "AND", "OR",
        "NOT", "BITWISE_AND", "BITWISE_OR", "BITWISE_LEFT",
        "BITWISE_RIGHT", "BITWISE_XOR", "BITWISE_NOT", "COMMENT", "MULTILINE_COMMENT",
        "INCLUDE", "RETURN", "ADDITION_ASSIGNMENT_OPERATOR", "SUBTRACTION_ASSIGNMENT_OPERATOR",
        "MULTIPLICATION_ASSIGNMENT_OPERATOR", "DIVISION_ASSIGNMENT_OPERATOR", "CONTINUE_CONTROL",
        "STRUCT", "INCREMENT", "DECREMENT", "NEWLINE", "NOP"
    ]

    literals = [
        '{', '}', '(', ')', ';', '=', ',', ':', '.', '[', ']'
    ]

    t_ignore = " \t\r"

    def t_NEWLINE(self, t):
        r'\n+'
        t.lexer.lineno += len(t.value)
        pass

    def t_NOP(self, t):
        r'nop'
        return t

    def t_COMMENT(self, t):
        r'//[^\n]*\n'
        t.lexer.lineno += 1
        pass

    def t_MULTILINE_COMMENT(self, t):
        r'\/\*(\*(?!\/)|[^*])*\*\/'
        t.lexer.lineno += t.value.count('\n')
        pass

    def t_INCLUDE(self, t):
        '\#include'
        return t

    def t_OBJECTSELF_VALUE(self, t):
        r'OBJECT_SELF'
        t.value = ObjectExpression(0)
        return t

    def t_OBJECTINVALID_VALUE(self, t):
        r'OBJECT_INVALID'
        t.value = ObjectExpression(1)
        return t

    def t_TRUE_VALUE(self, t):
        r'TRUE'
        t.value = IntExpression(1)
        return t

    def t_FALSE_VALUE(self, t):
        r'FALSE'
        t.value = IntExpression(0)
        return t

    # region Control Tokens
    def t_BREAK_CONTROL(self, t):
        r'break'
        t.value = ControlKeyword.BREAK
        return t

    def t_CONTINUE_CONTROL(self, t):
        r'continue'
        return t

    def t_CASE_CONTROL(self, t):
        r'case'
        t.value = ControlKeyword.CASE
        return t

    def t_DEFAULT_CONTROL(self, t):
        r'default'
        t.value = ControlKeyword.DEFAULT
        return t

    def t_DO_CONTROL(self, t):
        r'do'
        t.value = ControlKeyword.DO
        return t

    def t_ELSE_CONTROL(self, t):
        r'else'
        t.value = ControlKeyword.ELSE
        return t

    def t_SWITCH_CONTROL(self, t):
        r'switch'
        t.value = ControlKeyword.SWITCH
        return t

    def t_WHILE_CONTROL(self, t):
        r'while'
        t.value = ControlKeyword.WHILE
        return t

    def t_FOR_CONTROL(self, t):
        r'for'
        t.value = ControlKeyword.FOR
        return t

    def t_IF_CONTROL(self, t):
        r'if'
        t.value = ControlKeyword.IF
        return t

    def t_RETURN(self, t):
        r'return'
        t.value = ControlKeyword.RETURN
        return t
    # endregion

    # region Type Tokens
    def t_STRUCT(self, t):
        r'struct'
        t.value = DataType.STRUCT
        return t

    def t_INT_TYPE(self, t):
        r'int'
        t.value = DataType.INT
        return t

    def t_FLOAT_TYPE(self, t):
        r'float'
        t.value = DataType.FLOAT
        return t

    def t_OBJECT_TYPE(self, t):
        r'object'
        t.value = DataType.OBJECT
        return t

    def t_VOID_TYPE(self, t):
        r'void'
        t.value = DataType.VOID
        return t

    def t_EVENT_TYPE(self, t):
        r'event'
        t.value = DataType.EVENT
        return t

    def t_EFFECT_TYPE(self, t):
        r'effect'
        t.value = DataType.EFFECT
        return t

    def t_ITEMPROPERTY_TYPE(self, t):
        r'itemproperty'
        t.value = DataType.ITEMPROPERTY
        return t

    def t_LOCATION_TYPE(self, t):
        r'location'
        t.value = DataType.LOCATION
        return t

    def t_STRING_TYPE(self, t):
        r'string'
        t.value = DataType.STRING
        return t

    def t_TALENT_TYPE(self, t):
        r'talent'
        t.value = DataType.TALENT
        return t

    def t_ACTION_TYPE(self, t):
        r'action'
        t.value = DataType.ACTION
        return t

    def t_VECTOR_TYPE(self, t):
        r'vector'
        t.value = DataType.VECTOR
        return t
    # endregion

    # region Value Tokens
    def t_IDENTIFIER(self, t):
        r'[a-zA-Z_]+[a-zA-Z0-9_]*'
        t.value = Identifier(t.value)
        return t

    def t_STRING_VALUE(self, t):
        r'\"[^\"]*\"'
        t.value = StringExpression(t.value[1:-1])
        return t

    def t_FLOAT_VALUE(self, t):
        r'[0-9]+\.[0-9]+f?|[0-9]f'
        t.value = FloatExpression(float(t.value.replace("f", "")))
        return t

    def t_INT_HEX_VALUE(self, t):
        r'0x[0-9a-fA-F]+'
        t.value = IntExpression(int(t.value, 16))
        return t

    def t_INT_VALUE(self, t):
        r'[0-9]+'
        t.value = IntExpression(int(t.value))
        return t
    # endregion

    def t_INCREMENT(self, t):
        '\+\+'
        t.value = OperatorMapping([
            UnaryOperatorMapping(NCSInstructionType.INCISP, DataType.INT),
        ], [])
        return t

    def t_DECREMENT(self, t):
        '\-\-'
        t.value = OperatorMapping([
            UnaryOperatorMapping(NCSInstructionType.DECISP, DataType.INT),
        ], [])
        return t

    def t_ADDITION_ASSIGNMENT_OPERATOR(self, t):
        '\+\='
        return t

    def t_SUBTRACTION_ASSIGNMENT_OPERATOR(self, t):
        '\-\='
        return t

    def t_MULTIPLICATION_ASSIGNMENT_OPERATOR(self, t):
        '\*\='
        return t

    def t_DIVISION_ASSIGNMENT_OPERATOR(self, t):
        '/\='
        return t

    # region Operators
    def t_BITWISE_LEFT(self, t):
        '<<'
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.SHLEFTII, DataType.INT, DataType.INT, DataType.INT),
        ])
        return t

    def t_BITWISE_RIGHT(self, t):
        '>>'
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.SHRIGHTII, DataType.INT, DataType.INT, DataType.INT),
        ])
        return t

    def t_ADD(self, t):
        '\+'
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.ADDII, DataType.INT, DataType.INT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.ADDIF, DataType.INT, DataType.INT, DataType.FLOAT),
            BinaryOperatorMapping(NCSInstructionType.ADDFI, DataType.FLOAT, DataType.FLOAT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.ADDFF, DataType.FLOAT, DataType.FLOAT, DataType.FLOAT),
            BinaryOperatorMapping(NCSInstructionType.ADDVV, DataType.VECTOR, DataType.VECTOR, DataType.VECTOR),
            BinaryOperatorMapping(NCSInstructionType.ADDSS, DataType.STRING, DataType.STRING, DataType.STRING),
        ])
        return t

    def t_MINUS(self, t):
        '-'
        t.value = OperatorMapping([
            UnaryOperatorMapping(NCSInstructionType.NEGI, DataType.INT),
            UnaryOperatorMapping(NCSInstructionType.NEGF, DataType.FLOAT),
        ], [
            BinaryOperatorMapping(NCSInstructionType.SUBII, DataType.INT, DataType.INT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.SUBIF, DataType.INT, DataType.INT, DataType.FLOAT),
            BinaryOperatorMapping(NCSInstructionType.SUBFI, DataType.FLOAT, DataType.FLOAT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.SUBFF, DataType.FLOAT, DataType.FLOAT, DataType.FLOAT),
            BinaryOperatorMapping(NCSInstructionType.SUBVV, DataType.VECTOR, DataType.VECTOR, DataType.VECTOR),
        ])
        return t

    def t_MULTIPLY(self, t):
        '\*'
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.MULII, DataType.INT, DataType.INT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.MULIF, DataType.INT, DataType.INT, DataType.FLOAT),
            BinaryOperatorMapping(NCSInstructionType.MULFI, DataType.FLOAT, DataType.FLOAT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.MULFF, DataType.FLOAT, DataType.FLOAT, DataType.FLOAT),
            BinaryOperatorMapping(NCSInstructionType.MULVF, DataType.VECTOR, DataType.VECTOR, DataType.FLOAT),
            BinaryOperatorMapping(NCSInstructionType.MULFV, DataType.VECTOR, DataType.FLOAT, DataType.VECTOR),
        ])
        return t

    def t_DIVIDE(self, t):
        '/'
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.DIVII, DataType.INT, DataType.INT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.DIVIF, DataType.INT, DataType.INT, DataType.FLOAT),
            BinaryOperatorMapping(NCSInstructionType.DIVFI, DataType.FLOAT, DataType.FLOAT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.DIVFF, DataType.FLOAT, DataType.FLOAT, DataType.FLOAT),
            BinaryOperatorMapping(NCSInstructionType.DIVVF, DataType.VECTOR, DataType.VECTOR, DataType.FLOAT),
            BinaryOperatorMapping(NCSInstructionType.DIVFV, DataType.VECTOR, DataType.FLOAT, DataType.VECTOR),
        ])
        return t

    def t_MOD(self, t):
        '\%'
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.MODII, DataType.INT, DataType.INT, DataType.INT),
        ])
        return t

    def t_EQUALS(self, t):
        '\=\='
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.EQUALII, DataType.INT, DataType.INT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.EQUALFF, DataType.INT, DataType.FLOAT, DataType.FLOAT),
            BinaryOperatorMapping(NCSInstructionType.EQUALOO, DataType.INT, DataType.OBJECT, DataType.OBJECT),
            BinaryOperatorMapping(NCSInstructionType.EQUALSS, DataType.INT, DataType.STRING, DataType.STRING),
        ])
        return t

    def t_NOT_EQUALS(self, t):
        '\!='
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.NEQUALII, DataType.INT, DataType.INT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.NEQUALFF, DataType.INT, DataType.FLOAT, DataType.FLOAT),
            BinaryOperatorMapping(NCSInstructionType.NEQUALOO, DataType.INT, DataType.OBJECT, DataType.OBJECT),
            BinaryOperatorMapping(NCSInstructionType.NEQUALSS, DataType.INT, DataType.STRING, DataType.STRING),
        ])
        return t

    def t_GREATER_THAN_OR_EQUALS(self, t):
        '>\='
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.GEQII, DataType.INT, DataType.INT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.GEQFF, DataType.INT, DataType.FLOAT, DataType.FLOAT),
        ])
        return t

    def t_GREATER_THAN(self, t):
        '>'
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.GTII, DataType.INT, DataType.INT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.GTFF, DataType.INT, DataType.FLOAT, DataType.FLOAT),
        ])
        return t

    def t_LESS_THAN_OR_EQUALS(self, t):
        '\<='
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.LEQII, DataType.INT, DataType.INT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.LEQFF, DataType.INT, DataType.FLOAT, DataType.FLOAT),
        ])
        return t

    def t_LESS_THAN(self, t):
        '\<'
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.LTII, DataType.INT, DataType.INT, DataType.INT),
            BinaryOperatorMapping(NCSInstructionType.LTFF, DataType.INT, DataType.FLOAT, DataType.FLOAT),
        ])
        return t

    def t_AND(self, t):
        '&&'
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.LOGANDII, DataType.INT, DataType.INT, DataType.INT),
        ])
        return t

    def t_OR(self, t):
        '\|\|'
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.LOGORII, DataType.INT, DataType.INT, DataType.INT),
        ])
        return t

    def t_NOT(self, t):
        '\!'
        t.value = OperatorMapping([
            UnaryOperatorMapping(NCSInstructionType.NOTI, DataType.INT),
        ], [])
        return t

    def t_BITWISE_AND(self, t):
        '&'
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.BOOLANDII, DataType.INT, DataType.INT, DataType.INT),
        ])
        return t

    def t_BITWISE_OR(self, t):
        '\|'
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.INCORII, DataType.INT, DataType.INT, DataType.INT),
        ])
        return t

    def t_BITWISE_XOR(self, t):
        '\^'
        t.value = OperatorMapping([], [
            BinaryOperatorMapping(NCSInstructionType.EXCORII, DataType.INT, DataType.INT, DataType.INT),
        ])
        return t

    def t_BITWISE_NOT(self, t):
        '\~'
        t.value = OperatorMapping([UnaryOperatorMapping(NCSInstructionType.COMPI, DataType.INT)], [])
        return t
    # endregion



from abc import ABC, abstractmethod
from typing import List, Dict, Union


class PatcherMemory:
    def __init__(self):
        self.memory_2da: Dict[int, str] = {}
        self.memory_str: Dict[int, int] = {}  # StrRef# (token) -> dialog.tlk index


class TokenUsage(ABC):
    @abstractmethod
    def value(self, memory: PatcherMemory) -> str:
        ...


class NoTokenUsage(TokenUsage):
    def __init__(self, stored: Union[str, int]):
        self.stored = str(stored)

    def value(self, memory: PatcherMemory) -> str:
        return self.stored


class TokenUsage2DA(TokenUsage):
    def __init__(self, token_id: int):
        self.token_id: int = token_id

    def value(self, memory: PatcherMemory) -> str:
        return memory.memory_2da[self.token_id]


class TokenUsageTLK(TokenUsage):
    def __init__(self, token_id: int):
        self.token_id: int = token_id

    def value(self, memory: PatcherMemory) -> str:
        return str(memory.memory_str[self.token_id])



from __future__ import annotations

from pykotor.common.language import LocalizedString
from pykotor.common.misc import Game, ResRef
from pykotor.resource.formats.gff import GFF, GFFContent, read_gff, write_gff
from pykotor.resource.formats.gff.gff_auto import bytes_gff
from pykotor.resource.type import ResourceType, SOURCE_TYPES, TARGET_TYPES


class UTT:
    """
    Stores trigger data.

    Attributes:
        tag: "Tag" field.
        resref: "TemplateResRef" field.
        auto_remove_key: "AutoRemoveKey" field.
        faction_id: "Faction" field.
        cursor_id: "Cursor" field.
        highlight_height: "HighlightHeight" field.
        key_name: "KeyName" field.
        type_id: "Type" field.
        trap_detectable: "TrapDetectable" field.
        trap_detect_dc: "TrapDetectDC" field.
        trap_disarmable: "TrapDisarmable" field.
        trap_disarm_dc: "DisarmDC" field.
        is_trap: "TrapFlag" field.
        trap_once: "TrapOneShot" field.
        trap_type: "TrapType" field.
        on_disarm: "OnDisarm" field.
        on_trap_triggered: "OnTrapTriggered" field.
        on_click: "OnClick" field.
        on_heartbeat: "ScriptHeartbeat" field.
        on_enter: "ScriptOnEnter" field.
        on_exit: "ScriptOnExit" field.
        on_user_defined: "ScriptUserDefine" field.
        comment: "Comment" field.

        palette_id: "PaletteID" field. Used in toolset only.

        name: "LocalizedName" field. Not used by the game engine.
        loadscreen_id: "LoadScreenID" field. Not used by the game engine.
        portrait_id: "PortraitId" field. Not used by the game engine.
    """

    BINARY_TYPE = ResourceType.UTT

    def __init__(
            self
    ):
        self.resref: ResRef = ResRef.from_blank()
        self.comment: str = ""
        self.tag: str = ""

        self.name: LocalizedString = LocalizedString.from_invalid()

        self.faction_id: int = 0
        self.cursor_id: int = 0
        self.type_id: int = 0

        self.auto_remove_key: bool = True
        self.key_name: str = ""

        self.highlight_height: float = 0.0

        self.is_trap: bool = False
        self.trap_type: int = 0
        self.trap_once: bool = False
        self.trap_detectable: bool = False
        self.trap_detect_dc: int = 0
        self.trap_disarmable: bool = False
        self.trap_disarm_dc: int = 0

        self.on_disarm: ResRef = ResRef.from_blank()
        self.on_click: ResRef = ResRef.from_blank()
        self.on_trap_triggered: ResRef = ResRef.from_blank()
        self.on_heartbeat: ResRef = ResRef.from_blank()
        self.on_enter: ResRef = ResRef.from_blank()
        self.on_exit: ResRef = ResRef.from_blank()
        self.on_user_defined = ResRef.from_blank()

        # Deprecated:
        self.portrait_id: int = 0
        self.loadscreen_id: int = 0
        self.palette_id: int = 0
        self.name: LocalizedString = LocalizedString.from_invalid()


def construct_utt(
        gff: GFF
) -> UTT:
    utt = UTT()

    root = gff.root

    utt.tag = root.acquire("Tag", "")
    utt.resref = root.acquire("TemplateResRef", ResRef.from_blank())
    utt.auto_remove_key = root.acquire("AutoRemoveKey", 0)
    utt.faction_id = root.acquire("Faction", 0)
    utt.cursor_id = root.acquire("Cursor", 0)
    utt.highlight_height = root.acquire("HighlightHeight", 0.0)
    utt.key_name = root.acquire("KeyName", "")
    utt.type_id = root.acquire("Type", 0)
    utt.trap_detectable = root.acquire("TrapDetectable", 0)
    utt.trap_detect_dc = root.acquire("TrapDetectDC", 0)
    utt.trap_disarmable = root.acquire("TrapDisarmable", 0)
    utt.trap_disarm_dc = root.acquire("DisarmDC", 0)
    utt.is_trap = root.acquire("TrapFlag", 0)
    utt.trap_once = root.acquire("TrapOneShot", 0)
    utt.trap_type = root.acquire("TrapType", 0)
    utt.on_disarm = root.acquire("OnDisarm", ResRef.from_blank())
    utt.on_trap_triggered = root.acquire("OnTrapTriggered", ResRef.from_blank())
    utt.on_click = root.acquire("OnClick", ResRef.from_blank())
    utt.on_heartbeat = root.acquire("ScriptHeartbeat", ResRef.from_blank())
    utt.on_enter = root.acquire("ScriptOnEnter", ResRef.from_blank())
    utt.on_exit = root.acquire("ScriptOnExit", ResRef.from_blank())
    utt.on_user_defined = root.acquire("ScriptUserDefine", ResRef.from_blank())
    utt.comment = root.acquire("Comment", "")
    utt.name = root.acquire("LocalizedName", LocalizedString.from_invalid())
    utt.loadscreen_id = root.acquire("LoadScreenID", 0)
    utt.portrait_id = root.acquire("PortraitId", 0)
    utt.palette_id = root.acquire("PaletteID", 0)

    return utt


def dismantle_utt(
        utt: UTT,
        game: Game = Game.K2,
        *,
        use_deprecated: bool = True
) -> GFF:
    gff = GFF(GFFContent.UTT)

    root = gff.root
    root.set_string("Tag", utt.tag)
    root.set_resref("TemplateResRef", utt.resref)
    root.set_uint8("AutoRemoveKey", utt.auto_remove_key)
    root.set_uint32("Faction", utt.faction_id)
    root.set_uint8("Cursor", utt.cursor_id)
    root.set_single("HighlightHeight", utt.highlight_height)
    root.set_string("KeyName", utt.key_name)
    root.set_int32("Type", utt.type_id)
    root.set_uint8("TrapDetectable", utt.trap_detectable)
    root.set_uint8("TrapDetectDC", utt.trap_detect_dc)
    root.set_uint8("TrapDisarmable", utt.trap_disarmable)
    root.set_uint8("DisarmDC", utt.trap_disarm_dc)
    root.set_uint8("TrapFlag", utt.is_trap)
    root.set_uint8("TrapOneShot", utt.trap_once)
    root.set_uint8("TrapType", utt.trap_type)
    root.set_resref("OnDisarm", utt.on_disarm)
    root.set_resref("OnTrapTriggered", utt.on_trap_triggered)
    root.set_resref("OnClick", utt.on_click)
    root.set_resref("ScriptHeartbeat", utt.on_heartbeat)
    root.set_resref("ScriptOnEnter", utt.on_enter)
    root.set_resref("ScriptOnExit", utt.on_exit)
    root.set_resref("ScriptUserDefine", utt.on_user_defined)
    root.set_string("Comment", utt.comment)

    root.set_uint8("PaletteID", utt.palette_id)

    if use_deprecated:
        root.set_locstring("LocalizedName", utt.name)
        root.set_uint16("LoadScreenID", utt.loadscreen_id)
        root.set_uint16("PortraitId", utt.portrait_id)

    return gff


def read_utt(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> UTT:
    gff = read_gff(source, offset, size)
    utt = construct_utt(gff)
    return utt


def write_utt(
        utt: UTT,
        target: TARGET_TYPES,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> None:
    gff = dismantle_utt(utt, game, use_deprecated=use_deprecated)
    write_gff(gff, target, file_format)


def bytes_utt(
        utt: UTT,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> bytes:
    gff = dismantle_utt(utt, game, use_deprecated=use_deprecated)
    return bytes_gff(gff, file_format)



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from copy import deepcopy
from enum import Enum
from typing import List, Optional, Tuple, Dict, NamedTuple, Union, Any
from pykotor.common.script import DataType, ScriptFunction, ScriptConstant
from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.ncs import NCS, NCSInstruction, NCSInstructionType


def get_logical_equality_instruction(type1: DataType, type2: DataType) -> NCSInstructionType:
    if type1 == DataType.INT and type2 == DataType.INT:
        return NCSInstructionType.EQUALII
    elif type1 == DataType.FLOAT and type2 == DataType.FLOAT:
        return NCSInstructionType.EQUALFF
    elif type1 == DataType.FLOAT and type2 == DataType.FLOAT:
        return NCSInstructionType.EQUALSS
    elif type1 == DataType.FLOAT and type2 == DataType.FLOAT:
        return NCSInstructionType.EQUALOO
    else:
        raise CompileException(f"Tried an unsupported comparision between '{type1}' '{type2}'.")


class CompileException(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class TopLevelObject(ABC):
    @abstractmethod
    def compile(self, ncs: NCS, root: CodeRoot) -> None:
        ...


class GlobalVariableInitialization(TopLevelObject):
    def __init__(self, identifier: Identifier, data_type: DynamicDataType, value: Expression):
        super().__init__()
        self.identifier: Identifier = identifier
        self.data_type: DynamicDataType = data_type
        self.expression: Expression = value

    def compile(self, ncs: NCS, root: CodeRoot) -> None:
        block = CodeBlock()
        expression_type = self.expression.compile(ncs, root, block)
        if expression_type != self.data_type:
            raise CompileException(f"Tried to declare '{self.identifier}' a new variable with incorrect type '{expression_type}'.")
        root.add_scoped(self.identifier, self.data_type)


class GlobalVariableDeclaration(TopLevelObject):
    def __init__(self, identifier: Identifier, data_type: DynamicDataType):
        super().__init__()
        self.identifier: Identifier = identifier
        self.data_type: DynamicDataType = data_type

    def compile(self, ncs: NCS, root: CodeRoot) -> None:
        if self.data_type.builtin == DataType.INT:
            ncs.add(NCSInstructionType.RSADDI)
        elif self.data_type.builtin == DataType.FLOAT:
            ncs.add(NCSInstructionType.RSADDF)
        elif self.data_type.builtin == DataType.STRING:
            ncs.add(NCSInstructionType.RSADDS)
        elif self.data_type.builtin == DataType.OBJECT:
            ncs.add(NCSInstructionType.RSADDO)
        elif self.data_type.builtin == DataType.EVENT:
            ncs.add(NCSInstructionType.RSADDEVT)
        elif self.data_type.builtin == DataType.LOCATION:
            ncs.add(NCSInstructionType.RSADDLOC)
        elif self.data_type.builtin == DataType.TALENT:
            ncs.add(NCSInstructionType.RSADDTAL)
        elif self.data_type.builtin == DataType.EFFECT:
            ncs.add(NCSInstructionType.RSADDEFF)
        elif self.data_type.builtin == DataType.VECTOR:
            ncs.add(NCSInstructionType.RSADDF)
            ncs.add(NCSInstructionType.RSADDF)
            ncs.add(NCSInstructionType.RSADDF)
        elif self.data_type.builtin == DataType.STRUCT:
            root.struct_map[self.data_type._struct].initialize(ncs, root)
        elif self.data_type.builtin == DataType.VOID:
            raise CompileException("Cannot declare a variable of void type.")
        else:
            raise CompileException("Tried to compile a variable of unknown type.")

        root.add_scoped(self.identifier, self.data_type)


class Identifier:
    def __init__(self, label: str):
        self.label: str = label

    def __eq__(self, other):
        if isinstance(other, Identifier):
            return self.label == other.label
        elif isinstance(other, str):
            return self.label == other
        else:
            return NotImplemented

    def __str__(self):
        return self.label

    def __hash__(self):
        return hash(self.label)


class ControlKeyword(Enum):
    BREAK = "break"
    CASE = "control"
    DEFAULT = "default"
    DO = "do"
    ELSE = "else"
    SWITCH = "switch"
    WHILE = "while"
    FOR = "for"
    IF = "if"
    RETURN = "return"


class Operator(Enum):
    ADDITION = "+"
    SUBTRACT = "-"
    MULTIPLY = "*"
    DIVIDE = "/"
    MODULUS = "%"
    NOT = "!"
    EQUAL = "=="
    NOT_EQUAL = "!="
    GREATER_THAN = ">"
    LESS_THAN = "<"
    GREATER_THAN_OR_EQUAL = ">="
    LESS_THAN_OR_EQUAL = "<="
    AND = "&&"
    OR = "||"
    BITWISE_AND = "&"
    BITWISE_OR = "|"
    BITWISE_XOR = "^"
    BITWISE_LEFT = "<<"
    BITWISE_RIGHT = ">>"
    ONES_COMPLEMENT = "~"


class OperatorMapping(NamedTuple):
    unary: List[UnaryOperatorMapping]
    binary: List[BinaryOperatorMapping]


class BinaryOperatorMapping:
    def __init__(self, instruction: NCSInstructionType, result: DataType, lhs: DataType, rhs: DataType):
        self.instruction: NCSInstructionType = instruction
        self.result: DataType = result
        self.lhs: DataType = lhs
        self.rhs: DataType = rhs


class UnaryOperatorMapping:
    def __init__(self, instruction: NCSInstructionType, rhs: DataType):
        self.instruction: NCSInstructionType = instruction
        self.rhs: DataType = rhs


class FunctionReference(NamedTuple):
    instruction: NCSInstruction
    definition: Union[FunctionForwardDeclaration, FunctionDefinition]

    def is_prototype(self) -> bool:
        return isinstance(self.definition, FunctionForwardDeclaration)


class GetScopedResult(NamedTuple):
    isglobal: bool
    datatype: DynamicDataType
    offset: int


class Struct:
    def __init__(self, identifier: Identifier, members: List[StructMember]):
        self.identifier: Identifier = identifier
        self.members: List[StructMember] = members

    def initialize(self, ncs: NCS, root: CodeRoot) -> None:
        for member in self.members:
            member.initialize(ncs, root)

    def size(self, root: CodeRoot) -> int:
        # TODO: One-time calculation in __init__
        size = 0
        for member in self.members:
            size += member.size(root)
        return size

    def child_offset(self, root: CodeRoot, identifier: Identifier) -> int:
        """
        Returns the relative offset to the specified member inside the struct.
        """
        size = 0
        for member in self.members:
            if member.identifier == identifier:
                break
            size += member.size(root)
        else:
            raise CompileException(f"Trying to access unknown variable '{identifier}' on '{self.identifier}'.")
        return size

    def child_type(self, root: CodeRoot, identifier: Identifier) -> DynamicDataType:
        """
        Returns the child data type of the specified member inside the struct.
        """
        for member in self.members:
            if member.identifier == identifier:
                return member.datatype
        else:
            raise CompileException


class StructMember:
    def __init__(self, datatype: DynamicDataType, identifier: Identifier):
        self.datatype: DynamicDataType = datatype
        self.identifier: Identifier = identifier

    def initialize(self, ncs: NCS, root: CodeRoot) -> None:
        if self.datatype.builtin == DataType.INT:
            ncs.add(NCSInstructionType.RSADDI, args=[])
        elif self.datatype.builtin == DataType.FLOAT:
            ncs.add(NCSInstructionType.RSADDF, args=[])
        elif self.datatype.builtin == DataType.STRING:
            ncs.add(NCSInstructionType.RSADDS, args=[])
        elif self.datatype.builtin == DataType.OBJECT:
            ncs.add(NCSInstructionType.RSADDO, args=[])
        elif self.datatype.builtin == DataType.STRUCT:
            root.struct_map[self.identifier.label].initialize(ncs, root)
        else:
            raise CompileException

    def size(self, root: CodeRoot):
        return self.datatype.size(root)


class CodeRoot:
    def __init__(self, constants: List[ScriptConstant], functions: List[ScriptFunction], library_lookup: Optional[str],
                 library: Dict[str, bytes]):
        self.objects: List[TopLevelObject] = []

        self.library: Dict[str, bytes] = library
        self.functions: List[ScriptFunction] = functions
        self.constants: List[ScriptConstant] = constants
        self.library_lookup: Optional[str] = library_lookup

        self.function_map: Dict[str, FunctionReference] = {}
        self._global_scope: List[ScopedValue] = []
        self.struct_map: Dict[str, Struct] = {}

    def compile(self, ncs: NCS):
        # nwnnsscomp processes the includes and global variable declarations before functions regardless if they are
        # placed before or after function definitions. We will replicate this behaviour.

        included = []
        while [obj for obj in self.objects if isinstance(obj, IncludeScript)]:
            includes = [obj for obj in self.objects if isinstance(obj, IncludeScript)]
            include = includes.pop()
            self.objects.remove(include)
            included.append(include)
            include.compile(ncs, self)

        globals = [obj for obj in self.objects if isinstance(obj, GlobalVariableDeclaration) or isinstance(obj, GlobalVariableInitialization) or isinstance(obj, StructDefinition)]
        others = [obj for obj in self.objects if obj not in included and obj not in globals]

        if len(globals) > 0:
            for globaldef in globals:
                globaldef.compile(ncs, self)
            ncs.add(NCSInstructionType.SAVEBP, args=[])
        entry_index = len(ncs.instructions)

        for obj in others:
            obj.compile(ncs, self)

        if "main" in self.function_map:
            ncs.add(NCSInstructionType.RETN, args=[], index=entry_index)
            ncs.add(NCSInstructionType.JSR, jump=self.function_map["main"][0], index=entry_index)

    def compile_jsr(self, ncs: NCS, block: CodeBlock, name: str, *args: Expression) -> DynamicDataType:
        args = list(args)

        func_map = self.function_map[name]
        definition = func_map.definition
        start_instruction = func_map.instruction

        if definition.return_type == DynamicDataType.INT:
            ncs.add(NCSInstructionType.RSADDI, args=[])
        elif definition.return_type == DynamicDataType.FLOAT:
            ncs.add(NCSInstructionType.RSADDF, args=[])
        elif definition.return_type == DynamicDataType.STRING:
            ncs.add(NCSInstructionType.RSADDS, args=[])
        elif definition.return_type == DynamicDataType.VECTOR:
            raise NotImplementedError("Cannot define a function that returns Vector yet")  # TODO
        elif definition.return_type == DynamicDataType.OBJECT:
            ncs.add(NCSInstructionType.RSADDO, args=[])
        elif definition.return_type == DynamicDataType.TALENT:
            ncs.add(NCSInstructionType.RSADDTAL, args=[])
        elif definition.return_type == DynamicDataType.EVENT:
            ncs.add(NCSInstructionType.RSADDEVT, args=[])
        elif definition.return_type == DynamicDataType.LOCATION:
            ncs.add(NCSInstructionType.RSADDLOC, args=[])
        elif definition.return_type == DynamicDataType.EFFECT:
            ncs.add(NCSInstructionType.RSADDEFF, args=[])
        elif definition.return_type == DynamicDataType.VOID:
            ...
        else:
            raise NotImplementedError("Trying to return unsupported type?")  # TODO

        required_params = [param for param in definition.parameters if param.default is None]

        # Make sure the minimal number of arguments were passed through
        if len(required_params) > len(args):
            raise CompileException(f"Required argument missing in call to '{name}'.")

        # If some optional parameters were not specified, add the defaults to the arguments list
        while len(definition.parameters) > len(args):
            param_index = len(args)
            args.append(definition.parameters[param_index].default)

        offset = 0
        for param, arg in zip(definition.parameters, args):
            arg_datatype = arg.compile(ncs, self, block)
            offset += arg_datatype.size(self)
            block.tempstack += arg_datatype.size(self)
            if param.data_type != arg_datatype:
                raise CompileException(f"Wrong parameter type was past to '{param.identifier}' in function '{definition.identifier}'.")
        block.tempstack -= offset
        ncs.add(NCSInstructionType.JSR, jump=start_instruction)

        return definition.return_type

    def add_scoped(self, identifier: Identifier, datatype: DynamicDataType) -> None:
        self._global_scope.insert(0, ScopedValue(identifier, datatype))

    def get_scoped(self, identifier: Identifier, root: CodeRoot) -> GetScopedResult:
        offset = 0
        for scoped in self._global_scope:
            offset -= scoped.data_type.size(root)
            if scoped.identifier == identifier:
                break
        else:
            raise CompileException(f"Could not find variable '{identifier}'.")
        return GetScopedResult(True, scoped.data_type, offset)

    def scope_size(self):
        # TODO should be positive
        offset = 0
        for scoped in self._global_scope:
            offset -= scoped.data_type.size(self)
        return offset


class CodeBlock:
    def __init__(self):
        self.scope: List[ScopedValue] = []
        self._parent: Optional[CodeBlock] = None
        self._statements: List[Statement] = []
        self._break_scope: bool = False
        self.tempstack: int = 0

    def add(self, statement: Statement):
        self._statements.append(statement)

    def compile(self, ncs: NCS, root: CodeRoot, block: Optional[CodeBlock], return_instruction: NCSInstruction, break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        self._parent = block

        for statement in self._statements:
            if isinstance(statement, ReturnStatement):
                scope_size = self.full_scope_size(root)

                return_type = statement.compile(ncs, root, self, return_instruction, None, None)
                if return_type != DynamicDataType.VOID:
                    ncs.add(NCSInstructionType.CPDOWNSP, args=[-scope_size-return_type.size(root)*2, 4])
                    ncs.add(NCSInstructionType.MOVSP, args=[-return_type.size(root)])

                ncs.add(NCSInstructionType.MOVSP, args=[-scope_size])
                ncs.add(NCSInstructionType.JMP, jump=return_instruction)
                return
            else:
                statement.compile(ncs, root, self, return_instruction, break_instruction, continue_instruction)
        ncs.instructions.append(NCSInstruction(NCSInstructionType.MOVSP, [-self.scope_size(root)]))

        if self.tempstack != 0:
            # If the temp stack is 0 after the whole block has compiled there must be an logic error in the code
            # of one of the expression/statement classes
            raise ValueError

    def add_scoped(self, identifier: Identifier, data_type: DynamicDataType):
        self.scope.insert(0, ScopedValue(identifier, data_type))

    def get_scoped(self, identifier: Identifier, root: CodeRoot, offset: int = None) -> GetScopedResult:
        offset = -self.tempstack if offset is None else offset-self.tempstack
        for scoped in self.scope:
            offset -= scoped.data_type.size(root)
            if scoped.identifier == identifier:
                break
        else:
            if self._parent is not None:
                return self._parent.get_scoped(identifier, root, offset)
            else:
                return root.get_scoped(identifier, root)
        return GetScopedResult(False, scoped.data_type, offset)

    def scope_size(self, root: CodeRoot) -> int:
        """Returns size of local scope."""
        size = 0
        for scoped in self.scope:
            size += scoped.data_type.size(root)
        return size

    def full_scope_size(self, root: CodeRoot) -> int:
        """Returns size of scope, including outer blocks."""
        size = 0
        size += self.scope_size(root)
        if self._parent is not None:
            size += self._parent.full_scope_size(root)
        return size

    def break_scope_size(self, root: CodeRoot) -> int:
        """Returns size of scope up to the nearest loop/switch statement."""
        size = 0
        size += self.scope_size(root)
        if self._parent is not None and not self._parent._break_scope:
            size += self._parent.break_scope_size(root)
        return size

    def mark_break_scope(self) -> None:
        self._break_scope = True


class ScopedValue:
    def __init__(self, identifier: Identifier, data_type: DynamicDataType):
        self.identifier: Identifier = identifier
        self.data_type: DynamicDataType = data_type


class FunctionForwardDeclaration(TopLevelObject):
    def __init__(self, return_type: DynamicDataType, identifier: Identifier, parameters: List[FunctionDefinitionParam]):
        self.return_type: DynamicDataType = return_type
        self.identifier: Identifier = identifier
        self.parameters: List[FunctionDefinitionParam] = parameters

    def compile(self, ncs: NCS, root: CodeRoot):
        function_name = self.identifier.label

        if self.identifier.label in root.function_map:
            raise CompileException(f"Function '{function_name}' already has a prototype or been defined.")

        root.function_map[self.identifier.label] = FunctionReference(ncs.add(NCSInstructionType.NOP, args=[]), self)


class FunctionDefinition(TopLevelObject):
    # TODO: split definition into signature + block?
    def __init__(self, return_type: DynamicDataType, identifier: Identifier, parameters: List[FunctionDefinitionParam], block: CodeBlock):
        self.return_type: DynamicDataType = return_type
        self.identifier: Identifier = identifier
        self.parameters: List[FunctionDefinitionParam] = parameters
        self.block: CodeBlock = block

        for param in parameters:
            block.add_scoped(param.identifier, param.data_type)

    def compile(self, ncs: NCS, root: CodeRoot):
        name = self.identifier.label

        # Make sure all default parameters appear after the required parameters
        previous_is_default = False
        for param in self.parameters:
            is_default = param.default is not None
            if previous_is_default and not is_default:
                raise CompileException("Function parameter without a default value can't follow one with a default value.")
            previous_is_default = is_default

        # Make sure params are all constant values
        for param in self.parameters:
            if isinstance(param.default, IdentifierExpression) and not param.default.is_constant(root):
                raise CompileException(f"Non-constant default value specified for function prototype parameter '{param.identifier}'.")

        if name in root.function_map and not root.function_map[name].is_prototype():
            raise CompileException(f"Function '{name}' has already been defined.")
        elif name in root.function_map and root.function_map[name].is_prototype():

            if not self.is_matching_signature(root.function_map[name].definition):
                raise CompileException(f"Prototype of function '{name}' does not match definition.")

            # Function has forward declaration, insert the compiled definition after the stub
            temp = NCS()
            retn = NCSInstruction(NCSInstructionType.RETN)
            self.block.compile(temp, root, None, retn, None, None)
            temp.instructions.append(retn)

            stub_index = ncs.instructions.index(root.function_map[name].instruction)
            ncs.instructions[stub_index + 1:stub_index + 1] = temp.instructions
        else:
            retn = NCSInstruction(NCSInstructionType.RETN)

            function_start = ncs.add(NCSInstructionType.NOP, args=[])
            self.block.compile(ncs, root, None, retn, None, None)
            ncs.instructions.append(retn)

            root.function_map[name] = FunctionReference(function_start, self)

    def is_matching_signature(self, prototype: FunctionForwardDeclaration) -> bool:
        if self.return_type != prototype.return_type:
            return False
        if len(self.parameters) != len(prototype.parameters):
            return False
        for i in range(len(self.parameters)):
            if self.parameters[i].data_type != prototype.parameters[i].data_type:
                return False
            # TODO: nwnnsscomp compiles fine when default values do not match
            #       - how to handle? need some kind of warning system maybe.
            # if self.parameters[i].default != prototype.parameters[i].default:
            #     retrn False

        return True


class FunctionDefinitionParam:
    def __init__(self, data_type: DynamicDataType, identifier: Identifier, default: Optional[Expression] = None):
        self.data_type: DynamicDataType = data_type
        self.identifier: Identifier = identifier
        self.default: Optional[Expression] = default


class IncludeScript(TopLevelObject):
    def __init__(self, file: StringExpression, library: Dict[str, str] = None):
        self.file: StringExpression = file
        self.library: Dict[str, bytes] = library if library is not None else {}

    def compile(self, ncs: NCS, root: CodeRoot) -> None:
        lookup_filepath = os.path.normpath(f"{root.library_lookup}/{self.file.value}.nss")

        if root.library_lookup and os.path.exists(lookup_filepath):
            source = BinaryReader.load_file(lookup_filepath).decode(errors="ignore")
        elif self.file.value in self.library:
            source = self.library[self.file.value].decode(errors="ignore")
        else:
            raise CompileException(f"Could not find included script '{self.file.value}.nss'.")

        from pykotor.resource.formats.ncs.compiler.parser import NssParser
        nssParser = NssParser(root.functions, root.constants, root.library, root.library_lookup, None)
        nssParser.library = self.library
        nssParser.constants = root.constants
        t: CodeRoot = nssParser.parser.parse(source, tracking=True)
        root.objects = t.objects + root.objects


class StructDefinition(TopLevelObject):
    def __init__(self, identifier: Identifier, members: List[StructMember]):
        self.identifier: Identifier = identifier
        self.members: List[StructMember] = members

    def compile(self, ncs: NCS, root: CodeRoot) -> None:
        if len(self.members) == 0:
            raise CompileException("Struct cannot be empty.")
        root.struct_map[self.identifier.label] = Struct(self.identifier, self.members)


class Expression(ABC):
    @abstractmethod
    def compile(self, ncs: NCS, root: CodeRoot, block: Optional[CodeBlock]) -> DynamicDataType:
        ...


class Statement(ABC):
    def __init__(self):
        self.linenum: Optional[None] = None

    @abstractmethod
    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        ...


class FieldAccess:
    def __init__(self, identifiers: List[Identifier]):
        super().__init__()
        self.identifiers: List[Identifier] = identifiers

    def get_scoped(self, block: CodeBlock, root: CodeRoot) -> GetScopedResult:
        if len(self.identifiers) == 0:
            raise CompileException

        first = self.identifiers[0]
        scoped = block.get_scoped(first, root)

        isglobal = scoped.isglobal
        offset = scoped.offset
        datatype = scoped.datatype

        for next in self.identifiers[1:]:
            # Check previous datatype to see what members are accessible
            if datatype.builtin == DataType.VECTOR:
                datatype = DynamicDataType.FLOAT
                if next.label == "x":
                    offset += 0
                elif next.label == "y":
                    offset += 4
                elif next.label == "z":
                    offset += 8
                else:
                    raise CompileException(f"Attempting to access unknown member '{next}' on datatype '{datatype}'.")
            elif datatype.builtin == DataType.STRUCT:
                offset += root.struct_map[datatype._struct].child_offset(root, next)
                datatype = root.struct_map[datatype._struct].child_type(root, next)
            else:
                raise CompileException(f"Attempting to access unknown member '{next}' on datatype '{datatype}'.")

        return GetScopedResult(isglobal, datatype, offset)

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        isglobal, variable_type, stack_index = self.get_scoped(block, root)
        instruction_type = NCSInstructionType.CPTOPBP if isglobal else NCSInstructionType.CPTOPSP
        ncs.add(instruction_type, args=[stack_index, variable_type.size(root)])
        return variable_type


# region Expressions: Simple
class IdentifierExpression(Expression):
    def __init__(self, value: Identifier):
        super().__init__()
        self.identifier: Identifier = value

    def __eq__(self, other):
        if isinstance(other, IdentifierExpression):
            return self.identifier == other.identifier
        else:
            return NotImplemented

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        # Scan for any constants that are stored as part of the compiler (from nwscript).
        constant = next((constant for constant in root.constants if constant.name == self.identifier.label), None)
        if constant is not None:
            if constant.datatype == DataType.INT:
                ncs.add(NCSInstructionType.CONSTI, args=[int(constant.value)])
            elif constant.datatype == DataType.FLOAT:
                ncs.add(NCSInstructionType.CONSTF, args=[float(constant.value)])
            elif constant.datatype == DataType.STRING:
                ncs.add(NCSInstructionType.CONSTS, args=[str(constant.value)])
            return DynamicDataType(constant.datatype)
        else:
            isglobal, datatype, stack_index = block.get_scoped(self.identifier, root)
            instruction_type = NCSInstructionType.CPTOPBP if isglobal else NCSInstructionType.CPTOPSP
            ncs.add(instruction_type, args=[stack_index, datatype.size(root)])
            return datatype

    def is_constant(self, root: CodeRoot) -> bool:
        return next((constant for constant in root.constants if constant.name == self.identifier.label), None) is not None


class FieldAccessExpression(Expression):
    def __init__(self, field_access: FieldAccess):
        super().__init__()
        self.field_access: FieldAccess = field_access

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        scoped = self.field_access.get_scoped(block, root)
        instruction_type = NCSInstructionType.CPTOPBP if scoped.isglobal else NCSInstructionType.CPTOPSP
        ncs.instructions.append(NCSInstruction(instruction_type, [scoped.offset, scoped.datatype.size(root)]))
        return scoped.datatype


class StringExpression(Expression):
    def __init__(self, value: str):
        super().__init__()
        self.value: str = value

    def __eq__(self, other):
        if isinstance(other, StringExpression):
            return self.value == other.value
        else:
            return NotImplemented

    def data_type(self) -> DynamicDataType:
        return DynamicDataType.STRING

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        ncs.instructions.append(NCSInstruction(NCSInstructionType.CONSTS, [self.value]))
        return DynamicDataType.STRING


class IntExpression(Expression):
    def __init__(self, value: int):
        super().__init__()
        self.value: int = value

    def __eq__(self, other):
        if isinstance(other, IntExpression):
            return self.value == other.value
        else:
            return NotImplemented

    def data_type(self) -> DynamicDataType:
        return DynamicDataType.INT

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        ncs.instructions.append(NCSInstruction(NCSInstructionType.CONSTI, [self.value]))
        return DynamicDataType.INT


class ObjectExpression(Expression):
    def __init__(self, value: int):
        super().__init__()
        self.value: int = value

    def __eq__(self, other):
        if isinstance(other, ObjectExpression):
            return self.value == other.value
        else:
            return NotImplemented

    def data_type(self) -> DynamicDataType:
        return DynamicDataType.OBJECT

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        ncs.instructions.append(NCSInstruction(NCSInstructionType.CONSTO, [self.value]))
        return DynamicDataType.OBJECT


class FloatExpression(Expression):
    def __init__(self, value: float):
        super().__init__()
        self.value: float = value

    def __eq__(self, other):
        if isinstance(other, FloatExpression):
            return self.value == other.value
        else:
            return NotImplemented

    def data_type(self) -> DynamicDataType:
        return DynamicDataType.FLOAT

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        ncs.instructions.append(NCSInstruction(NCSInstructionType.CONSTF, [self.value]))
        return DynamicDataType.FLOAT


class VectorExpression(Expression):
    def __init__(self, x: FloatExpression, y: FloatExpression, z: FloatExpression):
        super().__init__()
        self.x: FloatExpression = x
        self.y: FloatExpression = y
        self.z: FloatExpression = z

    def __eq__(self, other):
        if isinstance(other, VectorExpression):
            return self.x == other.x and self.y == other.y and self.z == other.z
        else:
            return NotImplemented

    def data_type(self) -> DynamicDataType:
        return DynamicDataType.FLOAT

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        self.x.compile(ncs, root, block)
        self.y.compile(ncs, root, block)
        self.z.compile(ncs, root, block)
        return DynamicDataType.VECTOR


class EngineCallExpression(Expression):
    def __init__(self, function: ScriptFunction, routine_id: int, data_type: DynamicDataType, args: List[Expression]):
        super().__init__()
        self._function: ScriptFunction = function
        self._routine_id: int = routine_id
        self._args: List[Expression] = args

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        arg_count = len(self._args)

        if arg_count > len(self._function.params):
            raise CompileException(f"Too many arguments passed to '{self._function.name}'.")

        for i, param in enumerate(self._function.params):
            if i >= arg_count:
                if param.default is None:
                    raise CompileException(f"Not enough arguments passed to '{self._function.name}'.")
                else:
                    constant = next((constant for constant in root.constants if constant.name == param.default), None)
                    if constant is not None:
                        if constant.datatype == DataType.INT:
                            self._args.append(IntExpression(int(constant.value)))
                        elif constant.datatype == DataType.FLOAT:
                            self._args.append(FloatExpression(float(constant.value)))
                        elif constant.datatype == DataType.STRING:
                            self._args.append(StringExpression(str(constant.value)))
                        elif constant.datatype == DataType.OBJECT:
                            self._args.append(ObjectExpression(int(constant.value)))
                    else:
                        if param.datatype == DynamicDataType.INT:
                            self._args.append(IntExpression(int(param.default)))
                        elif param.datatype == DynamicDataType.FLOAT:
                            self._args.append(FloatExpression(float(param.default)))
                        elif param.datatype == DynamicDataType.STRING:
                            self._args.append(StringExpression(param.default))
                        elif param.datatype == DynamicDataType.VECTOR:
                            x = FloatExpression(param.default.x)
                            y = FloatExpression(param.default.y)
                            z = FloatExpression(param.default.z)
                            self._args.append(VectorExpression(x, y, z))
                        elif param.datatype == DynamicDataType.OBJECT:
                            self._args.append(ObjectExpression(int(param.default)))
                        else:
                            raise CompileException(f"Unexpected compilation error at '{self._function.name}' call.")

        this_stack = 0
        for i, arg in enumerate(reversed(self._args)):
            param_type = DynamicDataType(self._function.params[-i - 1].datatype)
            if param_type == DataType.ACTION:
                after_command = NCSInstruction()
                ncs.add(NCSInstructionType.STORE_STATE, args=[-root.scope_size(), block.full_scope_size(root)])
                ncs.add(NCSInstructionType.JMP, jump=after_command)
                arg.compile(ncs, root, block)
                ncs.instructions.append(after_command)
            else:
                added = arg.compile(ncs, root, block)
                block.tempstack += added.size(root)
                this_stack += added.size(root)

                if added != param_type:
                    raise CompileException(f"Tried to pass an argument of the incorrect type to '{self._function.name}'.")

        ncs.instructions.append(NCSInstruction(NCSInstructionType.ACTION, [self._routine_id, len(self._args)]))
        block.tempstack -= this_stack
        return DynamicDataType(self._function.returntype)


class FunctionCallExpression(Expression):
    def __init__(self, function: Identifier, args: List[Expression]):
        super().__init__()
        self._function: Identifier = function
        self._args: List[Expression] = args

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        if self._function.label not in root.function_map:
            raise CompileException(f"Function '{self._function.label}' has not been defined.")

        block.tempstack += root.function_map[self._function.label].definition.return_type.size(root)
        x = root.compile_jsr(ncs, block, self._function.label, *self._args)
        block.tempstack -= root.function_map[self._function.label].definition.return_type.size(root)
        return x
# endregion


class BinaryOperatorExpression(Expression):
    def __init__(self, expression1: Expression, expression2: Expression, mapping: List[BinaryOperatorMapping]):
        self.expression1: Expression = expression1
        self.expression2: Expression = expression2
        self.compatibility: List[BinaryOperatorMapping] = mapping

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        type1 = self.expression1.compile(ncs, root, block)
        block.tempstack += 4
        type2 = self.expression2.compile(ncs, root, block)
        block.tempstack += 4

        for x in self.compatibility:
            if type1 == x.lhs and type2 == x.rhs:
                ncs.add(x.instruction)
                break
        else:
            raise CompileException(f"Cannot compare {type1.builtin.name.lower()} against {type2.builtin.name.lower()}")

        block.tempstack -= 8
        return DynamicDataType(x.result)


class UnaryOperatorExpression(Expression):
    def __init__(self, expression1: Expression, mapping: List[UnaryOperatorMapping]):
        super().__init__()
        self.expression1: Expression = expression1
        self.compatibility: List[UnaryOperatorMapping] = mapping

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        type1 = self.expression1.compile(ncs, root, block)

        block.tempstack += 4

        for x in self.compatibility:
            if type1 == x.rhs:
                ncs.add(x.instruction)
                break
        else:
            raise CompileException(f"Cannot negate {type1.name.lower()}")

        block.tempstack -= 4
        return type1


class LogicalNotExpression(Expression):
    def __init__(self, expression1: Expression):
        super().__init__()
        self.expression1: Expression = expression1

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        type1 = self.expression1.compile(ncs, root, block)
        block.tempstack += 4

        if type1 == DynamicDataType.INT:
            ncs.add(NCSInstructionType.NOTI)
        else:
            raise CompileException(f"Cannot get the logical NOT of {type1.name.lower()}")

        block.tempstack -= 4
        return DynamicDataType.INT


class BitwiseNotExpression(Expression):
    def __init__(self, expression1: Expression):
        super().__init__()
        self.expression1: Expression = expression1

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        type1 = self.expression1.compile(ncs, root, block)
        block.tempstack += 4

        if type1 == DynamicDataType.INT:
            ncs.add(NCSInstructionType.COMPI)
        else:
            raise CompileException(f"Cannot get one's complement of {type1.name.lower()}")

        block.tempstack -= 4
        return type1


# region Expressions: Assignment
class Assignment(Expression):
    def __init__(self, field_access: FieldAccess, value: Expression):
        super().__init__()
        self.field_access: FieldAccess = field_access
        self.expression: Expression = value

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        variable_type = self.expression.compile(ncs, root, block)

        isglobal, expression_type, stack_index = self.field_access.get_scoped(block, root)
        instruction_type = NCSInstructionType.CPDOWNBP if isglobal else NCSInstructionType.CPDOWNSP
        stack_index -= 0 if isglobal else variable_type.size(root)

        if variable_type != expression_type:
            raise CompileException(f"Wrong type was assigned to symbol {self.field_access.identifiers[-1]}.")

        # Copy the value that the expression has already been placed on the stack to where the identifiers position is
        ncs.instructions.append(NCSInstruction(instruction_type, [stack_index, expression_type.size(root)]))

        return variable_type


class AdditionAssignment(Expression):
    def __init__(self, field_access: FieldAccess, value: Expression):
        super().__init__()
        self.field_access: FieldAccess = field_access
        self.expression: Expression = value

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        # Copy the variable to the top of the stack
        isglobal, variable_type, stack_index = self.field_access.get_scoped(block, root)
        instruction_type = NCSInstructionType.CPTOPBP if isglobal else NCSInstructionType.CPTOPSP
        ncs.add(instruction_type, args=[stack_index, variable_type.size(root)])
        block.tempstack += variable_type.size(root)

        # Add the result of the expression to the stack
        expresion_type = self.expression.compile(ncs, root, block)
        block.tempstack += expresion_type.size(root)

        # Determine what instruction to apply to the two values
        if variable_type == DynamicDataType.INT and expresion_type == DynamicDataType.INT:
            arthimetic_instruction = NCSInstructionType.ADDII
        elif variable_type == DynamicDataType.INT and expresion_type == DynamicDataType.FLOAT:
            arthimetic_instruction = NCSInstructionType.ADDIF
        elif variable_type == DynamicDataType.FLOAT and expresion_type == DynamicDataType.FLOAT:
            arthimetic_instruction = NCSInstructionType.ADDFF
        elif variable_type == DynamicDataType.FLOAT and expresion_type == DynamicDataType.INT:
            arthimetic_instruction = NCSInstructionType.ADDFI
        elif variable_type == DynamicDataType.STRING and expresion_type == DynamicDataType.STRING:
            arthimetic_instruction = NCSInstructionType.ADDSS
        else:
            raise CompileException(f"Wrong type was assigned to symbol {self.field_access.identifiers[-1]}.")

        # Add the expression and our temp variable copy together
        ncs.add(arthimetic_instruction, args=[])

        # Copy the result to the original variable in the stack
        ins_cpdown = NCSInstructionType.CPDOWNBP if isglobal else NCSInstructionType.CPDOWNSP
        offset_cpdown = stack_index if isglobal else stack_index - expresion_type.size(root)
        ncs.add(ins_cpdown, args=[offset_cpdown, variable_type.size(root)])

        block.tempstack -= variable_type.size(root)
        block.tempstack -= expresion_type.size(root)
        return expresion_type


class SubtractionAssignment(Expression):
    def __init__(self, field_access: FieldAccess, value: Expression):
        super().__init__()
        self.field_access: FieldAccess = field_access
        self.expression: Expression = value

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        # Copy the variable to the top of the stack
        isglobal, variable_type, stack_index = self.field_access.get_scoped(block, root)
        instruction_type = NCSInstructionType.CPTOPBP if isglobal else NCSInstructionType.CPTOPSP
        ncs.add(instruction_type, args=[stack_index, variable_type.size(root)])
        block.tempstack += variable_type.size(root)

        # Add the result of the expression to the stack
        expresion_type = self.expression.compile(ncs, root, block)
        block.tempstack += expresion_type.size(root)

        # Determine what instruction to apply to the two values
        if variable_type == DynamicDataType.INT and expresion_type == DynamicDataType.INT:
            arthimetic_instruction = NCSInstructionType.SUBII
        elif variable_type == DynamicDataType.INT and expresion_type == DynamicDataType.FLOAT:
            arthimetic_instruction = NCSInstructionType.SUBIF
        elif variable_type == DynamicDataType.FLOAT and expresion_type == DynamicDataType.FLOAT:
            arthimetic_instruction = NCSInstructionType.SUBFF
        elif variable_type == DynamicDataType.FLOAT and expresion_type == DynamicDataType.INT:
            arthimetic_instruction = NCSInstructionType.SUBFI
        else:
            raise CompileException(f"Wrong type was assigned to symbol {self.field_access.identifiers[-1]}.")

        # Add the expression and our temp variable copy together
        ncs.add(arthimetic_instruction)

        # Copy the result to the original variable in the stack
        ins_cpdown = NCSInstructionType.CPDOWNBP if isglobal else NCSInstructionType.CPDOWNSP
        offset_cpdown = stack_index if isglobal else stack_index - expresion_type.size(root)
        ncs.add(ins_cpdown, args=[offset_cpdown, variable_type.size(root)])

        block.tempstack -= expresion_type.size(root)
        block.tempstack -= variable_type.size(root)
        return expresion_type


class MultiplicationAssignment(Expression):
    def __init__(self, field_access: FieldAccess, value: Expression):
        super().__init__()
        self.field_access: FieldAccess = field_access
        self.expression: Expression = value

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        # Copy the variable to the top of the stack
        isglobal, variable_type, stack_index = self.field_access.get_scoped(block, root)
        instruction_type = NCSInstructionType.CPTOPBP if isglobal else NCSInstructionType.CPTOPSP
        ncs.add(instruction_type, args=[stack_index, variable_type.size(root)])
        block.tempstack += variable_type.size(root)

        # Add the result of the expression to the stack
        expresion_type = self.expression.compile(ncs, root, block)
        block.tempstack += expresion_type.size(root)

        # Determine what instruction to apply to the two values
        if variable_type == DynamicDataType.INT and expresion_type == DynamicDataType.INT:
            arthimetic_instruction = NCSInstructionType.MULII
        elif variable_type == DynamicDataType.INT and expresion_type == DynamicDataType.FLOAT:
            arthimetic_instruction = NCSInstructionType.MULIF
        elif variable_type == DynamicDataType.FLOAT and expresion_type == DynamicDataType.FLOAT:
            arthimetic_instruction = NCSInstructionType.MULFF
        elif variable_type == DynamicDataType.FLOAT and expresion_type == DynamicDataType.INT:
            arthimetic_instruction = NCSInstructionType.MULFI
        else:
            raise CompileException(f"Wrong type was assigned to symbol {self.field_access.identifiers[-1]}.")

        # Add the expression and our temp variable copy together
        ncs.add(arthimetic_instruction)

        # Copy the result to the original variable in the stack
        ins_cpdown = NCSInstructionType.CPDOWNBP if isglobal else NCSInstructionType.CPDOWNSP
        offset_cpdown = stack_index if isglobal else stack_index - expresion_type.size(root)
        ncs.add(ins_cpdown, args=[offset_cpdown, variable_type.size(root)])

        block.tempstack -= expresion_type.size(root)
        block.tempstack -= variable_type.size(root)
        return expresion_type


class DivisionAssignment(Expression):
    def __init__(self, field_access: FieldAccess, value: Expression):
        super().__init__()
        self.field_access: FieldAccess = field_access
        self.expression: Expression = value

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        # Copy the variable to the top of the stack
        isglobal, variable_type, stack_index = self.field_access.get_scoped(block, root)
        instruction_type = NCSInstructionType.CPTOPBP if isglobal else NCSInstructionType.CPTOPSP
        ncs.add(instruction_type, args=[stack_index, variable_type.size(root)])
        block.tempstack += variable_type.size(root)

        # Add the result of the expression to the stack
        expresion_type = self.expression.compile(ncs, root, block)
        block.tempstack += expresion_type.size(root)

        # Determine what instruction to apply to the two values
        if variable_type == DynamicDataType.INT and expresion_type == DynamicDataType.INT:
            arthimetic_instruction = NCSInstructionType.DIVII
        elif variable_type == DynamicDataType.INT and expresion_type == DynamicDataType.FLOAT:
            arthimetic_instruction = NCSInstructionType.DIVIF
        elif variable_type == DynamicDataType.FLOAT and expresion_type == DynamicDataType.FLOAT:
            arthimetic_instruction = NCSInstructionType.DIVFF
        elif variable_type == DynamicDataType.FLOAT and expresion_type == DynamicDataType.INT:
            arthimetic_instruction = NCSInstructionType.DIVFI
        else:
            raise CompileException(f"Wrong type was assigned to symbol {self.field_access.identifiers[-1]}.")

        # Add the expression and our temp variable copy together
        ncs.add(arthimetic_instruction)

        # Copy the result to the original variable in the stack
        ins_cpdown = NCSInstructionType.CPDOWNBP if isglobal else NCSInstructionType.CPDOWNSP
        offset_cpdown = stack_index if isglobal else stack_index - expresion_type.size(root)
        ncs.add(ins_cpdown, args=[offset_cpdown, variable_type.size(root)])

        block.tempstack -= expresion_type.size(root)
        block.tempstack -= variable_type.size(root)
        return expresion_type
# endregion


# region Statements
class EmptyStatement(Statement):
    def __init__(self):
        super().__init__()

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        return DynamicDataType.VOID


class NopStatement(Statement):
    def __init__(self, string: str):
        super().__init__()
        self.string = string

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        ncs.add(NCSInstructionType.NOP, args=[self.string])
        return DynamicDataType.VOID


class ExpressionStatement(Statement):
    def __init__(self, expression: Expression):
        super().__init__()
        self.expression: Expression = expression

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        expression_type = self.expression.compile(ncs, root, block)
        ncs.add(NCSInstructionType.MOVSP, args=[-expression_type.size(root)])


class DeclarationStatement(Statement):
    def __init__(self, data_type: DynamicDataType, declarators: List[VariableDeclarator]):
        super().__init__()
        self.data_type: DynamicDataType = data_type
        self.declarators: List[VariableDeclarator] = declarators

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        for declarator in self.declarators:
            declarator.compile(ncs, root, block, self.data_type)


class VariableDeclarator:
    def __init__(self, identifier: Identifier):
        self.identifier: Identifier = identifier

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, data_type: DynamicDataType):
        if data_type.builtin == DataType.INT:
            ncs.add(NCSInstructionType.RSADDI)
        elif data_type.builtin == DataType.FLOAT:
            ncs.add(NCSInstructionType.RSADDF)
        elif data_type.builtin == DataType.STRING:
            ncs.add(NCSInstructionType.RSADDS)
        elif data_type.builtin == DataType.OBJECT:
            ncs.add(NCSInstructionType.RSADDO)
        elif data_type.builtin == DataType.EVENT:
            ncs.add(NCSInstructionType.RSADDEVT)
        elif data_type.builtin == DataType.LOCATION:
            ncs.add(NCSInstructionType.RSADDLOC)
        elif data_type.builtin == DataType.TALENT:
            ncs.add(NCSInstructionType.RSADDTAL)
        elif data_type.builtin == DataType.EFFECT:
            ncs.add(NCSInstructionType.RSADDEFF)
        elif data_type.builtin == DataType.VECTOR:
            ncs.add(NCSInstructionType.RSADDF)
            ncs.add(NCSInstructionType.RSADDF)
            ncs.add(NCSInstructionType.RSADDF)
        elif data_type.builtin == DataType.STRUCT:
            root.struct_map[data_type._struct].initialize(ncs, root)
        elif data_type.builtin == DataType.VOID:
            raise CompileException("Cannot declare a variable of void type.")
        else:
            raise CompileException("Tried to compile a variable of unknown type.")

        block.add_scoped(self.identifier, data_type)


class VariableInitializer:
    def __init__(self, identifier: Identifier, expression: Expression):
        self.identifier: Identifier = identifier
        self.expression: Expression = expression

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, data_type: DynamicDataType):
        expression_type = self.expression.compile(ncs, root, block)
        if expression_type != data_type:
            raise CompileException(f"Tried to declare '{self.identifier}' a new variable with incorrect type '{expression_type.builtin}'.")
        block.add_scoped(self.identifier, data_type)


class ConditionalBlock(Statement):
    def __init__(self, if_block: ConditionAndBlock, else_if_blocks: List[ConditionAndBlock], else_block: CodeBlock):
        super().__init__()
        self.if_blocks: List[ConditionAndBlock] = [if_block] + else_if_blocks
        self.else_block: Optional[CodeBlock] = else_block

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        jump_count = 1 + len(self.if_blocks)
        jump_tos = [NCSInstruction(NCSInstructionType.NOP, args=[]) for i in range(jump_count)]

        for i, else_if in enumerate(self.if_blocks):
            else_if.condition.compile(ncs, root, block)
            ncs.add(NCSInstructionType.JZ, jump=jump_tos[i])

            else_if.block.compile(ncs, root, block, return_instruction, break_instruction, continue_instruction)
            ncs.add(NCSInstructionType.JMP, jump=jump_tos[-1])

            ncs.instructions.append(jump_tos[i])

        if self.else_block is not None:
            self.else_block.compile(ncs, root, block, return_instruction, break_instruction, continue_instruction)

        ncs.instructions.append(jump_tos[-1])

        '''self.condition.compile(ncs, root, block)
        jz = ncs.add(NCSInstructionType.JZ, jump=None)
        self.if_block.compile(ncs, root, block, return_instruction, break_instruction, continue_instruction)
        block_end = ncs.add(NCSInstructionType.NOP, args=[])

        # Set the Jump If Zero instruction to jump to the end of the block
        jz.jump = block_end'''


class ConditionAndBlock:
    def __init__(self, condition: Expression, block: CodeBlock):
        self.condition: Expression = condition
        self.block: CodeBlock = block


class ReturnStatement(Statement):
    def __init__(self, expression: Optional[Expression] = None):
        super().__init__()
        self.expression: Optional[Expression] = expression

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        if self.expression is not None:
            return self.expression.compile(ncs, root, block)
        return DynamicDataType.VOID


class WhileLoopBlock(Statement):
    def __init__(self, condition: Expression, block: CodeBlock):
        super().__init__()
        self.condition: Expression = condition
        self.block: CodeBlock = block

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        # Tell break/continue statements to stop here when getting scope size
        block.mark_break_scope()

        loopstart = ncs.add(NCSInstructionType.NOP, args=[])
        loopend = NCSInstruction(NCSInstructionType.NOP, args=[])
        condition_type = self.condition.compile(ncs, root, block)

        if condition_type != DynamicDataType.INT:
            raise CompileException("Condition must be int type.")

        ncs.add(NCSInstructionType.JZ, jump=loopend)
        self.block.compile(ncs, root, block, return_instruction, loopend, loopstart)
        ncs.add(NCSInstructionType.JMP, jump=loopstart)

        ncs.instructions.append(loopend)


class DoWhileLoopBlock(Statement):
    def __init__(self, condition: Expression, block: CodeBlock):
        super().__init__()
        self.condition: Expression = condition
        self.block: CodeBlock = block

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        # Tell break/continue statements to stop here when getting scope size
        block.mark_break_scope()

        loopstart = ncs.add(NCSInstructionType.NOP, args=[])
        conditionstart = NCSInstruction(NCSInstructionType.NOP, args=[])
        loopend = NCSInstruction(NCSInstructionType.NOP, args=[])

        self.block.compile(ncs, root, block, return_instruction, loopend, conditionstart)

        ncs.instructions.append(conditionstart)
        condition_type = self.condition.compile(ncs, root, block)
        if condition_type != DynamicDataType.INT:
            raise CompileException("Condition must be int type.")

        ncs.add(NCSInstructionType.JZ, jump=loopend)
        ncs.add(NCSInstructionType.JMP, jump=loopstart)
        ncs.instructions.append(loopend)


class ForLoopBlock(Statement):
    def __init__(self, initial: Expression, condition: Expression, iteration: Expression, block: CodeBlock):
        super().__init__()
        self.initial: Expression = initial
        self.condition: Expression = condition
        self.iteration: Expression = iteration
        self.block: CodeBlock = block

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        # Tell break/continue statements to stop here when getting scope size
        block.mark_break_scope()

        initial_type = self.initial.compile(ncs, root, block)
        ncs.add(NCSInstructionType.MOVSP, args=[-initial_type.size(root)])

        loopstart = ncs.add(NCSInstructionType.NOP, args=[])
        updatestart = NCSInstruction(NCSInstructionType.NOP, args=[])
        loopend = NCSInstruction(NCSInstructionType.NOP, args=[])

        condition_type = self.condition.compile(ncs, root, block)
        if condition_type != DynamicDataType.INT:
            raise CompileException("Condition must be int type.")

        ncs.add(NCSInstructionType.JZ, jump=loopend)
        self.block.compile(ncs, root, block, return_instruction, loopend, updatestart)

        ncs.instructions.append(updatestart)
        iteration_type = self.iteration.compile(ncs, root, block)
        ncs.add(NCSInstructionType.MOVSP, args=[-iteration_type.size(root)])

        ncs.add(NCSInstructionType.JMP, jump=loopstart)
        ncs.instructions.append(loopend)


class ScopedBlock(Statement):
    def __init__(self, block: CodeBlock):
        super().__init__()
        self.block: CodeBlock = block

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        self.block.compile(ncs, root, block, return_instruction, break_instruction, continue_instruction)
# endregion


class BreakStatement(Statement):
    def __init__(self):
        super().__init__()

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        if break_instruction is None:
            raise CompileException("Nothing to break out of.")
        ncs.add(NCSInstructionType.MOVSP, args=[-block.break_scope_size(root)])
        ncs.add(NCSInstructionType.JMP, jump=break_instruction)


class ContinueStatement(Statement):
    def __init__(self):
        super().__init__()

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        if continue_instruction is None:
            raise CompileException("Nothing to continue out of.")
        ncs.add(NCSInstructionType.MOVSP, args=[-block.break_scope_size(root)])
        ncs.add(NCSInstructionType.JMP, jump=continue_instruction)


class PrefixIncrementExpression(Expression):
    def __init__(self, field_access: FieldAccess):
        self.field_access: FieldAccess = field_access

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        variable_type = self.field_access.compile(ncs, root, block)

        if variable_type != DynamicDataType.INT:
            raise CompileException("Operator (++) not valid for specified types")

        isglobal, variable_type, stack_index = self.field_access.get_scoped(block, root)
        ncs.add(NCSInstructionType.INCISP, args=[-4])

        if isglobal:
            ncs.add(NCSInstructionType.CPDOWNBP, args=[stack_index, variable_type.size(root)])
        else:
            ncs.add(NCSInstructionType.CPDOWNSP, args=[stack_index - variable_type.size(root), variable_type.size(root)])

        return variable_type


class PostfixIncrementExpression(Expression):
    def __init__(self, field_access: FieldAccess):
        self.field_access: FieldAccess = field_access

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        variable_type = self.field_access.compile(ncs, root, block)
        block.tempstack += 4

        if variable_type != DynamicDataType.INT:
            raise CompileException("Operator (++) not valid for specified types")

        isglobal, variable_type, stack_index = self.field_access.get_scoped(block, root)
        if isglobal:
            ncs.add(NCSInstructionType.INCIBP, args=[stack_index])
        else:
            ncs.add(NCSInstructionType.INCISP, args=[stack_index])

        block.tempstack -= 4
        return variable_type


class PrefixDecrementExpression(Expression):
    def __init__(self, field_access: FieldAccess):
        self.field_access: FieldAccess = field_access

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        variable_type = self.field_access.compile(ncs, root, block)

        if variable_type != DynamicDataType.INT:
            raise CompileException("Operator (++) not valid for specified types")

        isglobal, variable_type, stack_index = self.field_access.get_scoped(block, root)
        ncs.add(NCSInstructionType.DECISP, args=[-4])

        if isglobal:
            ncs.add(NCSInstructionType.CPDOWNBP, args=[stack_index, variable_type.size(root)])
        else:
            ncs.add(NCSInstructionType.CPDOWNSP,
                    args=[stack_index - variable_type.size(root), variable_type.size(root)])

        return variable_type


class PostfixDecrementExpression(Expression):
    def __init__(self, field_access: FieldAccess):
        self.field_access: FieldAccess = field_access

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock) -> DynamicDataType:
        variable_type = self.field_access.compile(ncs, root, block)
        block.tempstack += 4

        if variable_type != DynamicDataType.INT:
            raise CompileException("Operator (++) not valid for specified types")

        isglobal, variable_type, stack_index = self.field_access.get_scoped(block, root)
        if isglobal:
            ncs.add(NCSInstructionType.DECIBP, args=[stack_index])
        else:
            ncs.add(NCSInstructionType.DECISP, args=[stack_index])

        block.tempstack -= 4
        return variable_type


# region Switch
class SwitchStatement(Statement):
    def __init__(self, expression: Expression, blocks: List[SwitchBlock]):
        super().__init__()
        self.expression: Expression = expression
        self.blocks: List[SwitchBlock] = blocks

        self.real_block: CodeBlock = CodeBlock()

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, return_instruction: NCSInstruction,
                break_instruction: Optional[NCSInstruction], continue_instruction: Optional[NCSInstruction]):
        self.real_block._parent = block
        block.mark_break_scope()

        block = self.real_block

        expression_type = self.expression.compile(ncs, root, block)
        block.tempstack += expression_type.size(root)

        end_of_switch = NCSInstruction(NCSInstructionType.NOP, args=[])

        tempncs = NCS()
        switchblock_to_instruction = {}
        for switchblock in self.blocks:
            switchblock_start = tempncs.add(NCSInstructionType.NOP, args=[])
            switchblock_to_instruction[switchblock] = switchblock_start
            for statement in switchblock.block:
                statement.compile(tempncs, root, block, return_instruction, end_of_switch, None)

        for switchblock in self.blocks:
            for label in switchblock.labels:
                # Do not want to run switch expression multiple times, execute it once and copy it to the top
                ncs.add(NCSInstructionType.CPTOPSP, args=[-expression_type.size(root), expression_type.size(root)])
                label.compile(ncs, root, block, switchblock_to_instruction[switchblock], expression_type)
        # If none of the labels match, jump over the code block
        ncs.add(NCSInstructionType.JMP, jump=end_of_switch)

        ncs.merge(tempncs)
        ncs.instructions.append(end_of_switch)

        # Pop the Switch expression
        ncs.add(NCSInstructionType.MOVSP, args=[-4])
        block.tempstack -= expression_type.size(root)


class SwitchBlock:
    def __init__(self, labels: List[SwitchLabel], block: List[Statement]):
        self.labels: List[SwitchLabel] = labels
        self.block: List[Statement] = block


class SwitchLabel(ABC):
    @abstractmethod
    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, jump_to: NCSInstruction, expression_type: DynamicDataType):
        ...


class ExpressionSwitchLabel:
    def __init__(self, expression: Expression):
        self.expression: Expression = expression

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, jump_to: NCSInstruction, expression_type: DynamicDataType):
        # Compare the copied Switch expression to the Label expression
        label_type = self.expression.compile(ncs, root, block)
        equality_instruction = get_logical_equality_instruction(expression_type, label_type)
        ncs.add(equality_instruction, args=[])

        # If the expressions match, then we jump to the appropriate place, otherwise continue trying the
        # other Labels
        ncs.add(NCSInstructionType.JNZ, jump=jump_to)


class DefaultSwitchLabel:
    def __init__(self):
        ...

    def compile(self, ncs: NCS, root: CodeRoot, block: CodeBlock, jump_to: NCSInstruction, expression_type: DynamicDataType):
        ncs.add(NCSInstructionType.JMP, jump=jump_to)
# endregion


class DynamicDataType:
    INT: DynamicDataType
    STRING: DynamicDataType
    FLOAT: DynamicDataType
    OBJECT: DynamicDataType
    VECTOR: DynamicDataType
    EVENT: DynamicDataType
    TALENT: DynamicDataType
    LOCATION: DynamicDataType
    EFFECT: DynamicDataType
    VOID: DynamicDataType

    def __init__(self, datatype: DataType, struct_name: str = None):
        self.builtin: DataType = datatype
        self._struct: str = struct_name

    def __eq__(self, other):
        if isinstance(other, DynamicDataType):
            if self.builtin == other.builtin:
                return self.builtin != DataType.STRUCT or (self.builtin == DataType.STRUCT and self._struct == other._struct)
            else:
                return False
        elif isinstance(other, DataType):
            return self.builtin == other and self.builtin != DataType.STRUCT
        else:
            return NotImplemented

    def size(self, root: CodeRoot) -> int:
        if self.builtin == DataType.STRUCT:
            return root.struct_map[self._struct].size(root)
        else:
            return self.builtin.size()


DynamicDataType.INT = DynamicDataType(DataType.INT)
DynamicDataType.STRING = DynamicDataType(DataType.STRING)
DynamicDataType.FLOAT = DynamicDataType(DataType.FLOAT)
DynamicDataType.OBJECT = DynamicDataType(DataType.OBJECT)
DynamicDataType.VECTOR = DynamicDataType(DataType.VECTOR)
DynamicDataType.VOID = DynamicDataType(DataType.VOID)
DynamicDataType.EVENT = DynamicDataType(DataType.EVENT)
DynamicDataType.TALENT = DynamicDataType(DataType.TALENT)
DynamicDataType.LOCATION = DynamicDataType(DataType.LOCATION)
DynamicDataType.EFFECT = DynamicDataType(DataType.EFFECT)
